(*Copyright (C) 2019, Daniel Wyckoff*)
(*This file is part of BooleanAlgebrasIntro2.

BooleanAlgebrasIntro2 is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

BooleanAlgebrasIntro2 is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BooleanAlgebrasIntro2.  If not,
see <http://www.gnu.org/licenses/>.*)
    
Require Import ProofIrrelevance.
Require Import FunctionalExtensionality.
Require Import NaryFunctions.
Require Import Equality.
Require Import TypeUtilities.
Require Import EqDec.
Require Import ArithUtilities.
Require Import FunctionProperties.
Require Import LogicUtilities.
Require Import DecidableDec.
Require Import ListUtilities.
Require Import SetUtilities.
Require Import Description.

(*Version 0.3 bool2 -- Certified -- Coq 8.4 pl4.*)

(*Is true when given list enumerates the elements in a FiniteT,
with respect to the structural recursion (and indexing).*)
Inductive FiniteT_list :
  forall {T}, FiniteT T -> list T -> Prop :=
| FiniteT_to_list_empty : FiniteT_list empty_finite nil
| FiniteT_to_list_add :
    forall {T} (pff:FiniteT T) (l:list T),
      FiniteT_list pff l ->
      FiniteT_list (add_finite T pff) (list_to_option l)
| FiniteT_to_list_bij :
    forall {T U} (pff:FiniteT T)
           (l:list T) (g:T->U)
           (pfinv:invertible g),
      FiniteT_list (bij_finite T U g pff pfinv) (map g l).


(*Finds minimum point at which f and g differ (with respect to an 
indexing..*)
Inductive disagrees_at_min {T U} {f g:T->U} 
          (pff:FiniteT T) (h:T->(FiniteT_card _ pff) @)
          (pffi:FiniteT_index pff h) :  
  disagrees f g -> T -> Prop :=
| disagrees_at_min_intro :
    forall c (pfn:f c <> g c),
      (forall c' (pfn':f c' <> g c'),
         proj1_sig (h c) <= proj1_sig (h c')) ->
      disagrees_at_min pff h pffi (disagrees_intro c f g pfn) c.


(*Do a sort on [nprod T] and a [FintieT_index] on it also.*)
Inductive lt_fun {T U}
          (pfft:FiniteT T) (ht:T->(FiniteT_card _ pfft) @)
          (pffu:FiniteT U) (hu:U->(FiniteT_card _ pffu) @)
          (pffit:FiniteT_index pfft ht)
          (pffiu:FiniteT_index pffu hu)
: Relation (T->U) :=
| lt_fun_intro : forall c (f g:T->U)
                        (pfd:disagrees f g),
                   disagrees_at_min pfft ht pffit pfd c ->
                   proj1_sig (hu (f c)) < proj1_sig (hu (g c)) ->
                   lt_fun pfft ht pffu hu pffit pffiu f g.



Definition transfer_sig_fun1
           {T U} {P Q:T->Prop} (pf:impl_pred P Q)
           (f:{x:T | P x} -> U) :
  {x:T | Q x} -> U :=
  fun y => f (transfer_sig pf y).

Definition transfer_sig_fun1'
           {T U} {P Q:T->Prop} (pf:iff_pred P Q)
           (f:{x:T | P x} -> U) :
  {x:T | Q x} -> U :=
  transfer_sig_fun1 (iff_pred2 _ _ pf) f.

Definition transfer_sig_fun2
           {T U} {P Q:U->Prop} (pf:pred_impl P Q)
           (f:T -> {y:U | P y}) :
  T -> {y:U | Q y} :=
  fun x => transfer_sig pf (f x).

Definition transfer_sig_fun2'
           {T U} {P Q:U->Prop} (pf:iff_pred P Q)
           (f:T -> {y:U | P y}) :
  T -> {y:U | Q y} :=
  transfer_sig_fun2 (iff_pred1 _ _ pf) f.

Definition transfer_sig_fun_dep
           {T U} {P:T->Prop} {Q R:U->Prop}
           (pf:pred_impl Q R)
           (f:prop_dep_fun P {y:U | Q y}) :
  prop_dep_fun P {y:U | R y} :=
  fun y (pfy:P y) =>
    let (x, pfx) := (f y pfy) in
    transfer_sig pf (exist Q x pfx).

Definition transfer_sig_fun_dep'
           {T U} {P:T->Prop} {Q R:U->Prop}
           (pf:iff_pred Q R)
           (f:prop_dep_fun P {y:U | Q y}) :
  prop_dep_fun P {y:U | R y} :=
  transfer_sig_fun_dep (iff_pred1 _ _ pf) f.


Lemma inj_transfer_sig_fun1 : 
  forall {T U} {P Q:T->Prop} (pf:impl_pred P Q)
         (f:{x:T | P x} -> U),
    Injective f ->
    Injective (transfer_sig_fun1 pf f).
  unfold transfer_sig_fun1, transfer_sig;
  intros ? ? ? ? ? ? h2; red in h2; red; intros ? ? h3;
  apply h2 in h3;
  apply exist_injective in h3;
  apply proj1_sig_injective in h3; auto.
Qed.

Lemma inj_transfer_sig_fun1' : 
  forall {T U} {P Q:T->Prop} (pf:iff_pred P Q)
         (f:{x:T | P x} -> U),
    Injective f ->
    Injective (transfer_sig_fun1' pf f).
  intros; apply inj_transfer_sig_fun1; auto.
Qed.

Lemma surj_transfer_sig_fun1' : 
  forall {T U} {P Q:T->Prop} (pf:iff_pred P Q)
         (f:{x:T | P x} -> U),
    surjective f ->
    surjective (transfer_sig_fun1' pf f).
  unfold transfer_sig_fun1', transfer_sig_fun1, transfer_sig.
  intros ? ? ? ? he f h1; red in h1; red; intro y; specialize (h1 y);
  destruct h1 as [x h2]; subst; destruct x as [x hx].
  red in he; destruct he as [he he'].
  red in he; do 2 red in he'.
  pose proof (he x hx) as h1.
  exists (exist Q x h1).
  repeat f_equal; apply proof_irrelevance.
Qed.

Lemma bij_transfer_sig_fun1' : 
  forall {T U} {P Q:T->Prop} (pf:iff_pred P Q)
         (f:{x:T | P x} -> U),
    bijective f ->
    bijective (transfer_sig_fun1' pf f).
  intros ? ? ? ? ? f h1; red in h1; red; destruct h1;split;
  [apply inj_transfer_sig_fun1'; auto |
   apply surj_transfer_sig_fun1'; auto].
Qed.

Lemma inv_transfer_sig_fun1' : 
  forall {T U} {P Q:T->Prop} (pf:iff_pred P Q)
         (f:{x:T | P x} -> U),
    invertible f ->
    invertible (transfer_sig_fun1' pf f).
  intros ? ? ? ? ? ? h1;
  apply invertible_impl_bijective in h1; 
  apply bijective_impl_invertible;
  apply bij_transfer_sig_fun1'; auto.
Qed.



Definition inj_transfer_sig_fun2 : 
  forall {T U} {P Q:U->Prop}
         (pf:pred_impl P Q)
         (f:T -> {y:U | P y}),
    Injective f ->
    Injective (transfer_sig_fun2 pf f).
  unfold transfer_sig_fun2, transfer_sig;
  intros ? ? ? ? ? ? h1; red in h1; red; intros ? ? h2;
  apply exist_injective in h2;
  apply proj1_sig_injective in h2; apply h1 in h2; auto.
Qed.

Definition inj_transfer_sig_fun2' : 
  forall {T U} {P Q:U->Prop}
         (pf:iff_pred P Q)
         (f:T -> {y:U | P y}),
    Injective f ->
    Injective (transfer_sig_fun2' pf f).
  intros; apply inj_transfer_sig_fun2; auto.
Qed.


Lemma surj_transfer_sig_fun2' :
  forall {T U} {P Q:U->Prop}
         (pf:iff_pred P Q)
         (f:T -> {y:U | P y}),
    surjective f ->
    surjective (transfer_sig_fun2' pf f).
  unfold
    transfer_sig_fun2',
    transfer_sig_fun2,
    transfer_sig;
  intros T U P Q h1 f h2; red in h2; red; intros x;
  destruct x as [x hx];
  destruct h1 as [h1 h3]; do 2 red in h3.
  pose proof (h3 x hx) as h4.
  specialize (h2 (exist _ _ h4)); destruct h2 as [y h5].
  exists y.
  apply proj1_sig_injective; auto; simpl.
  rewrite h5; auto.
Qed.

Lemma bij_transfer_sig_fun2' :
  forall {T U} {P Q:U->Prop}
         (pf:iff_pred P Q)
         (f:T -> {y:U | P y}),
    bijective f ->
    bijective (transfer_sig_fun2' pf f).
  intros ? ? ? ? ? ? h1; destruct h1; split;
  [apply inj_transfer_sig_fun2'; auto |
   apply surj_transfer_sig_fun2'; auto].
Qed.

Lemma invertible_transfer_sig_fun2' :
  forall {T U} {P Q:U->Prop}
         (pf:iff_pred P Q)
         (f:T -> {y:U | P y}),
    invertible f ->
    invertible (transfer_sig_fun2' pf f).
  intros ? ? ? ? ? ? h1;
  apply invertible_impl_bijective in h1; 
  apply bijective_impl_invertible;
  apply bij_transfer_sig_fun2'; auto.
Qed.


Lemma inj_transfer_sig_fun_dep :
  forall {T U} {P:T->Prop} {Q R:U->Prop}
         (pf:pred_impl Q R)
         (f:prop_dep_fun P {y:U | Q y}),
    inj_dep f ->
    inj_dep (transfer_sig_fun_dep pf f).
  unfold
    transfer_sig_fun_dep,
    transfer_sig;
  intros ? ? ? ? ? h1 f h2; red in h2; red;
  intros x y hx hy h5.
  specialize (h2 x y hx hy).
  destruct (f x hx), (f y hy).
  apply exist_injective in h5;
  apply proj1_sig_injective in h5;
  apply exist_injective in h5; subst.
  apply h2.
  f_equal.
  apply proof_irrelevance.
Qed.

Lemma inj_transfer_sig_fun_dep' :
  forall {T U} {P:T->Prop} {Q R:U->Prop}
         (pf:iff_pred Q R)
         (f:prop_dep_fun P {y:U | Q y}),
    inj_dep f ->
    inj_dep (transfer_sig_fun_dep' pf f).
  intros; apply inj_transfer_sig_fun_dep; auto.
Qed.

Lemma surj_transfer_sig_fun_dep' :
  forall {T U} {P:T->Prop} {Q R:U->Prop}
         (pf:iff_pred Q R)
         (f:prop_dep_fun P {y:U | Q y}),
    surj_dep f ->
    surj_dep (transfer_sig_fun_dep' pf f).
  unfold
    transfer_sig_fun_dep',
    transfer_sig_fun_dep,
    transfer_sig;
  intros ? ? ? ? ? he f h1; red in h1; red; intro y; destruct y as [y hy].
  destruct he as [h3 h4]; do 2 red in h4; pose proof (h4 _ hy) as h5.
  specialize (h1 (exist _ _ h5)); destruct h1 as [x [h1 h2]].
  exists x, h1.
  destruct (f x h1); apply exist_injective in h2; subst.
  f_equal; apply proof_irrelevance.
Qed.

Lemma bij_transfer_sig_fun_dep' :
  forall {T U} {P:T->Prop} {Q R:U->Prop}
         (pf:iff_pred Q R)
         (f:prop_dep_fun P {y:U | Q y}),
    bij_dep f ->
    bij_dep (transfer_sig_fun_dep' pf f).
  intros ? ? ? ? ? ? ? h1; red in h1; red; destruct h1 as [h1 h2]; split;
  [apply inj_transfer_sig_fun_dep'; auto |
   apply surj_transfer_sig_fun_dep'; auto].
Qed.

Lemma inv_id_prop_dep_sig' :
  forall {T:Type} (P:T->Prop),
    inv_dep (id_prop_dep_sig' P).
  unfold id_prop_dep_sig'; intros T P.
  gterm0.
  fold c.
  apply (inv_dep_intro c
                       (@proj1_sig T P)
                       (@proj2_sig T P)); unfold c; simpl; auto.
  intro y; destruct y; auto.
Qed.

Lemma inv_id_in_dep_sig' :
  forall {T:Type} (l:list T),
    inv_dep (id_in_dep_sig' l).
  intros; apply inv_id_prop_dep_sig'.
Qed.


Definition fun_subst {T U} (pfdt:type_eq_dec T) (f:T->U) (x:T) (y:U) : T -> U :=
  fun t => if (pfdt t x) then y else f x.

Definition seg_fun_transfer
           {U n n'} (pf:n = n') (f:seg_fun n U) : seg_fun n' U :=
  transfer_dep pf f (U := fun i => seg_fun i U).

Lemma seg_fun_transfer_eq_refl :
  forall {U n} (pf:n = n) (f:seg_fun n U),
    seg_fun_transfer pf f = f.
  unfold seg_fun_transfer; intros ? ? h1.
  pose proof (proof_irrelevance _ h1 (eq_refl _)); subst.
  intro; rewrite transfer_dep_eq_refl; auto.
Qed.

Lemma seg_fun_transfer_compat :
  forall {U n n'} (pfe:n = n')
         (f:seg_fun n U)
         i
         (pfi:i < n'),
    seg_fun_transfer pfe f i pfi =
    f i (lt_eq_trans _ _ _ pfi (eq_sym pfe)).
  intros; subst.
  rewrite seg_fun_transfer_eq_refl.
  f_equal; apply proof_irrelevance.
Qed.

Lemma seg_fun_transfer_eq :
  forall {U n n'} (pfe:n = n')
         (f:seg_fun n U),
    seg_fun_transfer pfe f =
    fun i (pfi:i < n') =>
      f i (lt_eq_trans _ _ _ pfi (eq_sym pfe)).
  intros; apply functional_extensionality_dep;
  intro; apply functional_extensionality; intro; 
  rewrite seg_fun_transfer_compat; auto.
Qed.


Lemma in_seg_fun_dec_transfer :
  forall {U n n'} (pfe:n = n')
         y (f:seg_fun n U),
    in_seg_fun_dec y f ->
    in_seg_fun_dec y (seg_fun_transfer pfe f).
  intros; subst; rewrite seg_fun_transfer_eq_refl; auto.
Qed.

Lemma in_seg_fun_dec_transfer' :
  forall {U n n'} (pfe:n = n')
         y (f:seg_fun n U),
    in_seg_fun_dec y (seg_fun_transfer pfe f) ->
    in_seg_fun_dec y f.
  intros ? ? ? ? ? ? h1;
  subst; rewrite seg_fun_transfer_eq_refl in h1; auto.
Qed.

Definition in_seg_fun_dec_subst
           {U n y y'} (f:seg_fun n U) (pfi:in_seg_fun_dec y f)
           (pfe:y = y') :
  in_seg_fun_dec y' f.
  intros; subst; auto.
Defined.

Lemma seg_fun_to_pred_compat :
   forall {U n} (f:seg_fun n U) i (pfi:i < pred n),
    seg_fun_to_pred f i pfi =
    f (S i) (lt_pred_S _ _ pfi).
  intros ? n;
  destruct n as [|n].
  intros; omega.
  intros f i.
  destruct i as [|i]; simpl; unfold seg_fun_to_S; intros; f_equal; apply proof_irrelevance.
Qed.



Lemma seg_fun_from_pred_compat :
  forall {U n} (f:seg_fun n U) i (pfi:i < pred n),
    seg_fun_from_pred f i pfi = f i (lt_pred' _ _ pfi).
  intros ? n;
  destruct n as [|n].
  intros; omega.
  intros f i.
  destruct i as [|i]; simpl; unfold seg_fun_from_S; intros; f_equal; apply proof_irrelevance.
Qed.


Definition seg_fun_dep_w {T n} (U:T->Type) (w:T^n) : Type :=
  forall i (pfi:i < n), U (nth_prod i pfi w).

Definition seg_fun_to_list {U n} := map_seg (U :=U ) (n := n).

Definition list_to_seg_fun {U} (l:list U) : seg_fun (length l) U :=
  fun i pf => nth_lt l i pf.

Lemma seg_fun_to_list_from_list :
  forall {U n} (hdu:type_eq_dec U) (f:seg_fun (S (S n)) U),
    seg_fun_to_list (seg_fun_to_S (seg_fun_from_S f)) =
    seg_fun_to_list (seg_fun_from_S (seg_fun_to_S f)).
  intros; rewrite leq_iff; auto.
  ex_goal. 
  do 2 rewrite length_map_seg; auto.
  exists hex.
  intros i hi h1.
  do 2 rewrite nth_lt_map_seg'.
  unfold seg_fun_to_S, seg_fun_from_S.
  f_equal.
  apply proof_irrelevance.
Qed.


Lemma seg_fun_to_list_to_S :
  forall {U n} (pfdu:type_eq_dec U)
         (f:seg_fun (S n) U),
    0 < n ->
    seg_fun_to_list (seg_fun_to_S f) = tl (seg_fun_to_list f).
  intros U n hdu; induction n as [|n ih]; simpl; auto.
  intros f ?.
  destruct (zerop n) as [|h1]; subst.
  simpl.
  unfold seg_fun_to_S; repeat f_equal; apply proof_irrelevance.
  specialize (ih (seg_fun_from_S f)).
  rewrite <- seg_fun_to_list_from_list.
  rewrite ih.
  simpl.
  rewrite tl_app.
  rewrite tl_app.
  rewrite tl_app.
  f_equal.
  unfold seg_fun_to_S; repeat f_equal; apply proof_irrelevance.
  rewrite map_seg_eq_nil_iff; omega.
  apply app_not_nil_r.
  intro; discriminate.
  rewrite map_seg_eq_nil_iff; omega.
  assumption.
  assumption.
Qed.

Fixpoint count_seg_fun_im
         {U n}
         (pfdu:type_eq_dec U) y :
  seg_fun n U -> nat :=
  match n with
    | 0 => fun _ => 0
    | S n' =>
      fun f  =>
        let f' := seg_fun_to_S f in
        let s := count_seg_fun_im pfdu y f' in
            if (pfdu y (f 0 (lt_0_Sn _))) then
              S s
            else
              s
  end.

Fixpoint count_seg_fun_im_nat
         {n} y :
  seg_fun n nat -> nat :=
  match n with
    | 0 => fun _ => 0
    | S n' =>
      fun f  =>
        let f' := seg_fun_to_S f in
        let s := count_seg_fun_im_nat y f' in
            if (beq_nat y (f 0 (lt_0_Sn _))) then
              S s
            else
              s
  end.

Lemma count_seg_fun_im_functional :
  forall {U n} (pfdu:type_eq_dec U) y y'
         (f f':seg_fun n U),
    y = y' ->
    f = f' ->
    count_seg_fun_im pfdu y f =
    count_seg_fun_im pfdu y' f'.
  intros; subst; auto.
Qed.

Lemma count_seg_fun_im_nat_functional :
  forall {n}  y y'
         (f f':seg_fun n nat),
    y = y' ->
    f = f' ->
    count_seg_fun_im nat_eq_dec y f =
    count_seg_fun_im nat_eq_dec y' f'.
  intros; subst; auto.
Qed.



Lemma unq_inv_dep_ex :
  forall {T U:Type} {P:T->Prop} 
         (f:prop_dep_fun P U)
         (pfinv:inv_dep f),
  exists! pr:{g:U->T & (forall u, P (g u))},
      (forall (x:T) (pfx:P x), (projT1 pr) (f x pfx) = x) /\
      (forall (y:U), f ((projT1 pr) y) (projT2 pr y) = y).
  intros T U P f hinv. 
  induction hinv. 
  exists (existT _ g pfp); red; simpl.
  split; try intros g' hg'; auto.
  destruct g' as [g' h2]; simpl in hg'.
  eapply projT_injective; simpl.
  apply proof_irrelevance.

  Grab Existential Variables.
  simpl.
  apply functional_extensionality.
  intro y.
  destruct hg' as [h3 h4].
  specialize (H0 y).
  specialize (h4 y).
  rewrite <- h4 at 1.
  rewrite H; auto.
Qed.



Fixpoint nth_seg_fun_prod_type {n} :
  forall {f:seg_fun n Type}
         (g:seg_fun_prod_type f) 
         i pfi, 
    f i pfi :=
  match n with
    | 0 => fun _ _ _ h1  => False_rect _ (lt_n_0 _ h1)
    | S n' => 
      fun f g =>
        let f' := seg_fun_to_S f in
        let (fg, sg) := g in
        fun i =>
          match i with
            | 0 => fun pf0 => transfer_dep_pf (lt_0_Sn _) pf0 fg
                                              (P := fun i => i < S n')
                                              (U := fun j pfj => f j pfj)
                                              (x := 0)
            | S i' => fun pfi =>
                        let pfi' := lt_S_n _ _ pfi in
                        transfer_seg_fun_to_S_app f i' pfi' pfi 
                          (@nth_seg_fun_prod_type
                            n' f' sg i' pfi')
          end
  end.


Inductive in_seg_fun_prod_type_aux {n} 
          (f:seg_fun n Type)
          (g:seg_fun_prod_type f) 
          i (pfi:i < n) (x:f i pfi) : Prop :=
| in_seg_fun_prod_type_inro : 
    forall j (pfj:j < n)
           (pfe:f j pfj = f i pfi),
      transfer pfe (nth_seg_fun_prod_type g j pfj) = x ->
      in_seg_fun_prod_type_aux f g i pfi x.



Definition in_seg_fun_prod_type
           {n} 
           {f:seg_fun n Type}
           {i} {pfi:i < n}
           (x:f i pfi)
           (g:seg_fun_prod_type f) : Prop :=
  in_seg_fun_prod_type_aux f g i pfi x.


Lemma seg_fun_prod_type_to_ex :
  forall {n} {f:seg_fun n Type}
         (g:seg_fun_prod_type f),
  exists! (k:seg_fun_dep f),
    g = seg_fun_prod k.
  intro n.
  induction n as [|n ih]; simpl.
  intros f x.
  exists (fun i (pfi:i < 0) => False_rect _ (lt_n_0 _ pfi)).
  red.
  split.
  apply unit_eq.
  intros f' h1; subst.
  apply functional_extensionality_dep.
  intro.
  apply functional_extensionality_dep.
  intro; omega.
  intros f g.
  destruct g as [fg sg].
  specialize (ih (seg_fun_to_S f) sg).
  destruct ih as [k ih].
  exists (fun i =>
            match i with
              | 0 =>
                fun (pf0:0 < S n) => transfer_dep_pf (lt_0_Sn n) pf0 fg
                                                     (x:=0)
              | S i' =>
                fun (pfi:S i' < S n) =>
                  let pfi' := lt_S_pred' _ _ (lt_0_Sn _) pfi in
                  transfer_seg_fun_to_S_app f i' pfi' pfi (k i' pfi')
            end).
  red.
  split.
  f_equal.
  unfold transfer_dep_pf. 
  rewrite transfer_dep_eq_refl; auto.
  red in ih.
  destruct ih as [h1 h2]; subst.
  f_equal.
  apply functional_extensionality_dep.
  intro i.
  apply functional_extensionality_dep.
  intro hi.
  unfold seg_fun_dep_to_S.
  pose proof (transfer_seg_fun_to_S_app_subst
                k i hi) as hki.
  rewrite <- hki.
  pose proof (proof_irrelevance
                _ hi
                (lt_S_pred' (S i) n (lt_0_Sn i) (lt_n_S i n hi))) as h3.
  rewrite <- h3; auto.

  red in ih.
  destruct ih as [h5 h6].
  subst.
     
  intros sg' h1.
  inversion h1; subst.
  inversion h1.
  specialize (h6 _ H0); subst.
  apply functional_extensionality_dep.
  intro i.
  apply functional_extensionality_dep.
  intro hi.
  destruct i as [|i].
  unfold transfer_dep_pf.
  pose proof (proof_irrelevance _ (lt_0_Sn _) hi) as h2.
  rewrite <- h2.
  rewrite eq_eq_refl at 1.
  rewrite transfer_dep_eq_refl; auto.
  unfold transfer_seg_fun_to_S_app,
  eq_rect_r, eq_rect.
  pose proof (proof_irrelevance
                _
                (lt_n_S i n (lt_S_pred' (S i) n (lt_0_Sn i) hi))
                hi) as h2.
  match goal with
      |- context [match ?pf with | _ => _ end] => pose pf as h3 end.
  pose (lt_S_pred' (S i) n (lt_0_Sn i) hi) as hlt.
  fold hlt in h2.
  pose proof (lt_S_n _ _ hi) as h4.
  pose proof (proof_irrelevance _ hlt h4) as h5.
  fold hlt.
  rewrite h5.
  unfold hlt in h2.
  fold hlt in h2.
  rewrite h5 in h2.
  clear h5. clear hlt.
  subst.
  rewrite eq_eq_refl at 1.
  unfold seg_fun_dep_to_S; auto.
Qed.


Definition seg_fun_prod_type_to 
           {n} {f:seg_fun n Type}
           (g:seg_fun_prod_type f) :
  seg_fun_dep f :=
  proj1_sig (constructive_definite_description
               _
               (seg_fun_prod_type_to_ex g)).


Lemma seg_fun_prod_type_to_compat :
  forall {n} {f:seg_fun n Type}
         (g:seg_fun_prod_type f),
    let k := seg_fun_prod_type_to g in
    g = seg_fun_prod k.
  unfold seg_fun_prod_type_to; intros;
  destruct constructive_definite_description; auto.
Qed.

  
Definition inverse_dep 
           {T U:Type} {P:T->Prop} 
           (f:prop_dep_fun P U)
           (pfinv:inv_dep f) :
  {g:U->T & (forall u, P (g u))} :=
  proj1_sig (constructive_definite_description _ (unq_inv_dep_ex f pfinv)).

Lemma inverse_dep_compat :  
  forall {T U:Type} {P:T->Prop} 
         (f:prop_dep_fun P U)
         (pfinv:inv_dep f),
    let pr := inverse_dep f pfinv in
  (forall (x:T) (pfx:P x), (projT1 pr) (f x pfx) = x) /\
  (forall (y:U), f ((projT1 pr) y) (projT2 pr y) = y).
  unfold inverse_dep; intros; destruct constructive_definite_description; auto.
Qed.

Definition inverse_dep'
           {T U:Type} {P:T->Prop} 
           (f:prop_dep_fun P U)
           (pfinv:inv_dep f) :
  U->{x:T | P x} :=
  let pr := inverse_dep f pfinv in
  fun y => exist P (projT1 pr y) (projT2 pr y).

Lemma inverse_dep_compat' :  
  forall {T U:Type} {P:T->Prop} 
         (f:prop_dep_fun P U)
         (pfinv:inv_dep f),
    let g := inverse_dep' f pfinv in
    (forall (x:T) (pfx:P x), g (f x pfx) = exist _ x pfx) /\
    (forall (y:U), f (proj1_sig (g y)) (proj2_sig (g y)) = y).
  unfold inverse_dep'; simpl; intros ? ? ? f hinv;
  pose proof (inverse_dep_compat f hinv) as h1; destruct h1;
  split; intro; auto.
  intro; apply proj1_sig_injective; simpl; auto.
Qed.

Lemma inverse_dep_compat'' :
  forall {T U} {P:T->Prop} (f:prop_dep_fun P U)
         (pfinv:inv_dep f) x,
    proj1_sig (inverse_dep' f pfinv x) = 
    proj1_sig (inverse_dep f pfinv) x.
  intros ? ? ? f hinv x; unfold inverse_dep'; simpl.
  pose proof (inverse_dep_compat' f hinv) as h1; simpl in h1.
  destruct h1 as [h1 h3].
  specialize (h3 x).
  destruct (inverse_dep f hinv); simpl; auto.
Qed.

Lemma bij_impl_inv_dep :
  forall {T U:Type} {P:T->Prop} (f:prop_dep_fun P U),
    bij_dep f ->
    inv_dep f.
  intros T U P f h1. 
  red in h1.
  destruct h1 as [h1 h2].
  red in h1, h2.
  assert (g:forall y:U, {x:T | exists pfx:P x, y = f x pfx}).
    intro.
    apply constructive_definite_description.
    specialize (h2 y).
    destruct h2 as [x [h2 h3]]; subst.
    exists x.
    red.
    split.
    exists h2; auto.
    intros x' h3.
    destruct h3 as [h3 h4].
    eapply h1.
    apply h4. 
  pose proof (inv_dep_intro f (fun y:U => proj1_sig (g y))
                            (fun y:U =>
                               (destruct_exists_pf_set_aux _ _ _ (proj2_sig (g y))))) as h3.
  apply h3; clear h3.
  intros x hx.
  destruct (g (f x hx)); simpl.
  destruct e as [h3 h4].
  eapply h1.
  apply (eq_sym h4).
  intro y.
  destruct (g y); simpl.
  destruct e as [h3 h4]; subst.
  f_equal; apply proof_irrelevance.
Qed.

Lemma inv_impl_bij_dep :
  forall {T U:Type} {P:T->Prop} (f:prop_dep_fun P U),
    inv_dep f ->
    bij_dep f.
  intros T U P f h1.
  pose proof (inverse_dep_compat f h1) as h2; simpl in h2.
  destruct h2 as [h2 h3].
  constructor; red.
  intros x y hx hy h4.
  pose proof (h2 _ hx) as h5.
  pose proof (h2 _ hy) as h6.
  rewrite <- h4 in h6.
  congruence.
  intro y.
  specialize (h3 y).
  exists (projT1 (inverse_dep f h1) y), (projT2 (inverse_dep f h1) y); auto.
Qed.

Lemma inv_impl_inj_dep :
  forall {T U:Type} {P:T->Prop} (f:prop_dep_fun P U),
    inv_dep f ->
    inj_dep f.
  intros ? ? ? ? h1; apply inv_impl_bij_dep in h1; destruct h1; auto.
Qed.

Lemma inv_impl_surj_dep :
  forall {T U:Type} {P:T->Prop} (f:prop_dep_fun P U),
    inv_dep f ->
    surj_dep f.
  intros ? ? ? ? h1; apply inv_impl_bij_dep in h1; destruct h1; auto.
Qed.

Lemma inverse_dep_inv' :
  forall {T U:Type} {P:T->Prop} 
         (f:prop_dep_fun P U)
         (pfinv:inv_dep f),
    invertible (inverse_dep' f pfinv).
  intros T U P f hinv.
  pose proof (inverse_dep_compat' f hinv) as h4; simpl in h4;
  destruct h4 as [h4 h5].
  pose proof hinv as h1.
  apply inv_impl_bij_dep in h1.
  apply bijective_impl_invertible.
  red in h1; red; destruct h1 as [h1 h2]; split; red.
  intros x y h3.
  pose proof (h2 x) as h6.
  destruct h6 as [a [h6 ?]]; subst.
  pose proof (h2 y) as h7.
  destruct h7 as [b [h7 ?]]; subst.
  do 2 rewrite h4 in h3.
  apply exist_injective in h3; subst; f_equal; apply proof_irrelevance.
  intro y.
  destruct y as [y hy].
  exists (f y hy); auto.
Qed.

Lemma inverse_dep_bij' :
  forall {T U:Type} {P:T->Prop} 
         (f:prop_dep_fun P U)
         (pfinv:inv_dep f),
    bijective (inverse_dep' f pfinv).
  intros; apply invertible_impl_bijective; apply inverse_dep_inv'.
Qed.

Lemma inverse_dep_inj' :
  forall {T U:Type} {P:T->Prop} 
         (f:prop_dep_fun P U)
         (pfinv:inv_dep f),
    Injective (inverse_dep' f pfinv).
  intros; apply invertible_impl_injective; apply inverse_dep_inv'.
Qed.

Lemma inverse_dep_surj' :
  forall {T U:Type} {P:T->Prop} 
         (f:prop_dep_fun P U)
         (pfinv:inv_dep f),
    surjective (inverse_dep' f pfinv).
  intros; apply invertible_impl_surjective; apply inverse_dep_inv'.
Qed.


Lemma inv_fun_in_dep_fun_compose :
  forall {T U V:Type} {P:T->Prop}
         (f:prop_dep_fun P U)
         (g:U->V),
    inv_dep f ->
    invertible g ->
    inv_dep (fun_in_dep_fun_compose f g).
  intros T U V P f g h1 h2.
  apply bij_impl_inv_dep.
  apply inv_impl_bij_dep in h1.
  apply invertible_impl_bijective in h2.
  red in h1, h2.
  destruct h1 as [h1a h1b], h2 as [h2a h2b].
  red; split; red.
  intros x y hx hy h1.
  unfold fun_in_dep_fun_compose in h1.
  apply h2a in h1.
  apply h1a in h1; auto.
  intro y.
  specialize (h2b y).
  destruct h2b as [x]; subst.
  specialize (h1b x).
  destruct h1b as [z [hz ?]]; subst.
  exists z, hz.
  unfold fun_in_dep_fun_compose; auto.
Qed.

Lemma inv_tl :
  forall {T U} {l:list T}
         (f:fun_in_dep l U)
         (pf0:l <> nil)
         (pfi:In (hd' l pf0) (tl l)),
    inv_dep f ->
    inv_dep (tl_in_dep f).
  intros T U l f h0 hi h1.
  apply inv_impl_bij_dep in h1.
  apply bij_impl_inv_dep.
  eapply bij_tl; auto.
  apply hi.
Qed.


Lemma inv_dep_sig :
  forall {T U} {P:T->Prop} (f:prop_dep_fun P U),
    inv_dep f->
    invertible (prop_dep_fun_sig f).
  unfold prop_dep_fun_sig; intros T U P f h1.
  apply bijective_impl_invertible.
  apply inv_impl_bij_dep in h1.
  apply bij_dep_sig; auto.
Qed.

Lemma inv_transfer_sig_fun_dep' :
  forall {T U} {P:T->Prop} {Q R:U->Prop}
         (pf:iff_pred Q R)
         (f:prop_dep_fun P {y:U | Q y}),
    inv_dep f ->
    inv_dep (transfer_sig_fun_dep' pf f).
  unfold
    transfer_sig_fun_dep',
  transfer_sig_fun_dep; intros T U P Q R h1 f h2;
  apply inv_impl_bij_dep in h2;
  apply bij_impl_inv_dep;
  auto;
  apply bij_transfer_sig_fun_dep'; auto.
Qed.


Lemma false_dep_eq :
  forall {T U} (f:T->False->U),
    inv_dep f ->
    equi U False.
  intros T U f h1. 
  pose proof (inv_dep_sig _ h1) as h2.
  pose proof (invertible_impl_inv_invertible _ h2) as h4. 
  eapply equi_trans.
  econstructor.
  apply h4.
  apply equi_sym.
  apply equi_sig_false.
Qed.

Lemma equi_false_inv :
  forall T,
    equi T False ->
    invertible (false_fun T).
  intros T h1.
  apply bijective_impl_invertible.
  inversion h1.
  red; split; red.
  intros; contradiction.
  intro y.
  pose (f y); contradiction.
Qed.



Lemma surj_false_fun_false :
  forall U,
    ~ (equi U False) ->
    surjective (false_fun (U->False)).
  intro; red; intros h1 f.
  pose proof (inv_fun_false f) as h2.
  contradict h1.
  econstructor.
  apply h2.
Qed.

Lemma bij_false_fun_false :
  forall U,
    ~ (equi U False) ->
    bijective (false_fun (U->False)).
  intros; red; split;
  [apply inj_false_fun_false | apply surj_false_fun_false]; auto.
Qed.

Lemma inv_false_fun_false :
  forall U,
    ~ (equi U False) ->
    invertible (false_fun (U->False)).
  intros;
  apply bijective_impl_invertible; apply bij_false_fun_false; auto.
Qed.


Lemma equi_fun_false_false :
  forall U,
    ~ equi U False ->
    equi (U->False) False.
  intros U h1.
  pose proof (inv_false_fun_false _ h1) as h2.
  apply equi_sym.
  econstructor.
  apply h2.
Qed.


         
Definition count_seg_fun {U n} (pfdu:type_eq_dec U) 
           (f:seg_fun n U) (i:nat) (pfi:i < n) : nat :=
  count_seg_fun_im pfdu (f i pfi) f.

Definition count_seg_fun_nat {n} (f:seg_fun n nat) (i:nat)
           (pfi:i < n) : nat :=
  count_seg_fun_im_nat (f i pfi) f.


Lemma count_seg_fun_im_compat :
  forall {U n} (pfdu:type_eq_dec U)
         (f:seg_fun n U)
         i (pfi:i < n),
    count_seg_fun_im pfdu (f i pfi) f =
    count_seg_fun pfdu f i pfi.
  auto.
Qed.

Lemma count_seg_fun_im_nat_compat :
  forall {n} 
         (f:seg_fun n nat)
         i (pfi:i < n),
    count_seg_fun_im_nat (f i pfi) f =
    count_seg_fun_nat f i pfi.
  auto.
Qed.


Lemma count_seg_fun_im_nat_compat' :
  forall {n} (f:seg_fun n nat) y,
    count_seg_fun_im_nat y f =
    count_seg_fun_im nat_eq_dec y f.
  intro n.
  induction n as [|n ih]; simpl; auto.
  intros f y.
  destruct (nat_eq_dec y (f 0 (lt_0_Sn n))) as [|h1]; subst.
  rewrite beq_nat_same.
  f_equal; auto.
  rewrite <- neq_nat_compat in h1.
  rewrite h1; auto.
Qed.


Lemma count_seg_fun_nat_compat :
  forall {n} (f:seg_fun n nat),
    count_seg_fun_nat f =
    count_seg_fun nat_eq_dec f.
  intros n f; apply functional_extensionality_dep; intro i;
  apply functional_extensionality; intro hi.
  unfold count_seg_fun, count_seg_fun_nat.
  apply count_seg_fun_im_nat_compat'.
Qed.


Lemma count_seg_fun_im_to_S :
  forall {U n} (pfdu:type_eq_dec U)
         (f:seg_fun (S n) U),
    count_seg_fun_im pfdu (f 0 (lt_0_Sn _)) f =
    S (count_seg_fun_im pfdu (f 0 (lt_0_Sn _)) (seg_fun_to_S f)).
  intros; simpl; lr_if_goal'; auto.
  contradict hr; auto.
Qed.

Lemma count_seg_fun_im_in_iff :
  forall {n U} (pfdu:type_eq_dec U) (f:seg_fun n U) y,
    count_seg_fun_im pfdu y f > 0 <->
    in_seg_fun y f.
  intro n.
  induction n as [|n ih]; simpl.
  intuition.
  intros U hdu f y.
  lr_if_goal; fold hlr; destruct hlr as [hl | hr]; subst.
  split; intro h1; auto with arith.
  split; intro h1.
  rewrite ih in h1.
  right; auto.
  destruct h1; subst.
  contradict hr; auto.
  rewrite ih; auto.
Qed.


Lemma count_seg_fun_im_in_nat_iff :
  forall {n} (f:seg_fun n nat) y,
    count_seg_fun_im_nat y f > 0 <->
    in_seg_fun y f.
  intros n f y. 
  pose proof (count_seg_fun_im_nat_compat' f y) as h1.
  rewrite h1.
  apply count_seg_fun_im_in_iff.
Qed.

Lemma count_seg_fun_im_eq0_iff :
  forall {U n} (pfdu:type_eq_dec U) (f:seg_fun n U) y,
    count_seg_fun_im pfdu y f = 0 <->
    ~in_seg_fun y f.
  intros; rewrite <- count_seg_fun_im_in_iff. 
  split; intro h1; [rewrite h1 at 1 | omega]; auto with arith.
Qed. 

Lemma count_seg_fun_im_nat_eq0_iff :
  forall {n} (f:seg_fun n nat) y,
    count_seg_fun_im_nat y f = 0 <->
    ~in_seg_fun y f.
  intros; rewrite count_seg_fun_im_nat_compat';
  apply count_seg_fun_im_eq0_iff.
Qed.


Lemma count_seg_fun_im2 :
  forall {U n} (pfdu:type_eq_dec U)
         (f:seg_fun n U)
         i j (pfi:i < n) (pfj:j < n),
    i <> j ->
    f i pfi = f j pfj ->
    count_seg_fun_im pfdu (f i pfi) f > 1.
  intros U n.
  induction n as [|n ih]; simpl.
  intros; omega .
  intros hdu f i j hi hj hn he.
  destruct i as [|i].
  erewrite f_equal_dep.
  rewrite eq_dec_same at 1.
  apply lt_n_S.
  apply gt_lt. 
  rewrite count_seg_fun_im_in_iff.
  erewrite f_equal_dep.
  rewrite he at 1.
  rewrite in_seg_fun_iff.
  exists (pred j).
  ex_goal.
  omega.
  exists hex.
  unfold seg_fun_to_S.
  gen0.
  apply neq_sym in hn.
  rewrite <- (S_pred _ _ (iff1 (neq0_iff _) hn)).
  intro; f_equal; apply proof_irrelevance.
  reflexivity.
  assumption.
  reflexivity.
  destruct j as [|j].
  rewrite he.
  erewrite f_equal_dep.
  rewrite eq_dec_same at 1.
  apply lt_n_S.
  apply gt_lt.
  rewrite count_seg_fun_im_in_iff.
  erewrite f_equal_dep.
  rewrite <- he at 1.
  erewrite f_equal_dep.
  erewrite seg_fun_to_S_compat'.
  apply in_seg_fun_compat.
  reflexivity.
  reflexivity.
  assumption.
  reflexivity.
  lr_if_goal'.
  apply lt_n_S.
  apply gt_lt.
  rewrite count_seg_fun_im_in_iff.
  erewrite f_equal_dep.
  erewrite seg_fun_to_S_compat' at 1.
  apply in_seg_fun_compat.
  reflexivity.
  specialize (ih hdu (seg_fun_to_S f) i j
                 (lt_S_n _ _ hi)
                 (lt_S_n _ _ hj)).  
  hfirst ih.
  contradict hn; subst; auto.
  specialize (ih hf).
  hfirst ih. 
  unfold seg_fun_to_S. 
  erewrite f_equal_dep. 
  rewrite he at 1. 
  f_equal; apply proof_irrelevance.
  reflexivity.
  specialize (ih hf0).
  unfold seg_fun in ih.
  erewrite f_equal_dep.
  apply ih. 
  reflexivity.

  Grab Existential Variables.
  auto with arith.
  auto with arith.
Qed.

(*Returns a strict least upper bound of a [seg_fun] into [nat]*)
Fixpoint seg_fun_lub_weak {n} : seg_fun n nat -> nat :=
  match n with
    | 0 => fun _ => 0
    | S n' => fun f =>
                let f' := seg_fun_to_S f in
                let b := seg_fun_lub_weak f' in
                if (ge_nat b (f 0 (lt_0_Sn _))) then
                  b
                else
                  f 0 (lt_0_Sn _)
  end.

Definition seg_fun_lub {n} (f:seg_fun n nat) : nat :=
  S (seg_fun_lub_weak f).

Lemma le_seg_fun_lub_weak0 :
  forall {n} (pf:0 < n)
         (f:seg_fun n nat),
  f 0 pf <= seg_fun_lub_weak f. 
  intro n. 
  induction n as [|n ih]; simpl.
  intros; omega.
  intros h1 f.
  destruct n as [|n]; simpl.
  destruct (zerop (f 0 (lt_0_Sn _))) as [h2 | h2].
  rewrite h2.
  rewrite <- h2.
  erewrite f_equal_dep.
  apply le_refl.
  reflexivity.
  rewrite (S_pred _ _ h2).
  rewrite <- (S_pred _ _ h2).
  erewrite f_equal_dep.
  apply le_refl.
  reflexivity.
  destruct (lt_le_dec (seg_fun_lub_weak (seg_fun_to_S (seg_fun_to_S f)))
                      (seg_fun_to_S f 0 (lt_0_Sn n))) as [h5 | h5].
  pose proof h5 as h5'.
  apply lt_not_le in h5'.
  rewrite  <- le_nat_compat, le_nat_ge, <- eq_false_iff in h5'.
  rewrite h5'.
  destruct (lt_le_dec (seg_fun_to_S f 0 (lt_0_Sn n)) (f 0 (lt_0_Sn (S n)))) as [h6 | h6].
  pose proof h6 as h6'.
  apply lt_not_le in h6'.
  rewrite  <- le_nat_compat, le_nat_ge, <- eq_false_iff in h6'.
  rewrite h6'.
  erewrite f_equal_dep.
  apply le_refl.
  reflexivity.
  pose proof h6 as h6'.
  rewrite  <- le_nat_compat, le_nat_ge in h6'.
  rewrite h6'.
  erewrite f_equal_dep.
  apply h6.
  reflexivity.
  pose proof h5 as h5'.
  rewrite  <- le_nat_compat, le_nat_ge in h5'.
  rewrite h5'.
  
  destruct (lt_le_dec (seg_fun_lub_weak (seg_fun_to_S (seg_fun_to_S f)))
                      (f 0 (lt_0_Sn (S n)))) as [h4 | h4].
  pose proof h4 as h4'.
  apply lt_not_le in h4'.
  rewrite  <- le_nat_compat, le_nat_ge, <- eq_false_iff in h4'.
  rewrite h4'.
  erewrite f_equal_dep.
  apply le_refl.
  reflexivity.
  pose proof h4 as h4'.
  rewrite <- le_nat_compat, le_nat_ge in h4'.
  rewrite h4'.
  specialize (ih (lt_0_Sn _) (seg_fun_to_S f)).
  eapply le_trans.
  erewrite f_equal_dep.
  apply h4.
  reflexivity.
  apply le_refl.
Qed.


Lemma lt_seg_fun_lub0 :
  forall {n} (pf:0 < n)
         (f:seg_fun n nat),
    f 0 pf < seg_fun_lub f.
  unfold seg_fun_lub; intros; apply le_lt; apply le_seg_fun_lub_weak0.
Qed.


Lemma seg_fun_lub_compat :
  forall {n} (f:seg_fun n nat),
    0 < n ->
    (forall i (pfi:i < n),
       f i pfi < seg_fun_lub f)
     /\
     (forall b,
        (forall i (pfi:i < n),
           f i pfi < b) -> 
        seg_fun_lub f <= b).
  unfold seg_fun_lub; intro n. 
  induction n as [|n ih]; simpl.
  intros ? h1.
  contradict (lt_irrefl _ h1).
  intro f; split. 
  intros i h1.
  destruct i as [|i]. 
  destruct (lt_le_dec (seg_fun_lub_weak (seg_fun_to_S f))
                      (f 0 (lt_0_Sn n))) as [h4 | h4].
  pose proof h4 as h4'.
  apply lt_not_le in h4'.
  rewrite  <- le_nat_compat, le_nat_ge, <- eq_false_iff in h4'.
  rewrite h4'.
  erewrite f_equal_dep.
  apply lt_n_Sn.
  reflexivity. 
  pose proof h4 as h4'.
  rewrite <- le_nat_compat, le_nat_ge in h4.
  rewrite h4.
  apply le_lt in h4'.
  erewrite f_equal_dep.
  apply h4'.
  reflexivity.
  pose proof (lt_S_n _ _ h1) as h2.
  specialize (ih (seg_fun_to_S f) (Olt _ _ h2)).
  destruct ih as [h3 h4].
  specialize (h3 i h2).
  destruct (lt_le_dec (seg_fun_lub_weak (seg_fun_to_S f)) (f 0 (lt_0_Sn n))) as [h5 | h5].
  pose proof h5 as h5'.
  apply lt_gt in h5' .
  apply lt_not_le in h5'.
  apply nle_nge in h5'.
  rewrite <- ge_nat_compat in h5'.
  rewrite <- eq_false_iff in h5'.
  rewrite h5'.
  unfold seg_fun_to_S in h3 at 1.
  eapply lt_le_trans.
  erewrite f_equal_dep.
  apply h3.
  reflexivity.
  apply le_n_S.
  auto with arith.
  apply le_ge in h5.
  rewrite <- ge_nat_compat in h5.
  rewrite h5.
  unfold seg_fun_to_S in h3 at 1.
  erewrite f_equal_dep.
  apply h3.
  reflexivity.
  intros b h1.
  destruct (lt_le_dec (seg_fun_lub_weak (seg_fun_to_S f))
                      (f 0 (lt_0_Sn n))) as [h2 | h2].
  pose proof h2 as h3.
  apply lt_not_le in h3.
  apply nle_nge in h3.
  rewrite <- ge_nat_compat in h3.
  rewrite <- eq_false_iff in h3.
  rewrite h3.
  apply lt_le.
  apply lt_n_S; auto.
  apply le_ge in h2.
  pose proof h2 as h2'.
  rewrite <- ge_nat_compat in h2.
  rewrite h2.
  apply lt_le.
  apply lt_n_S; auto.
  destruct (zerop n) as [|h3]; subst; simpl.
  specialize (h1 0 (lt_0_Sn _)); omega.
  specialize (ih (seg_fun_to_S f) h3).
  destruct ih as [h4 h5].
  specialize (h5 b).
  hfirst h5.
  intros i hi.
  unfold seg_fun_to_S.
  specialize (h1 _ (lt_n_S _ _ hi)); auto.
  specialize (h5 hf).
  omega.
Qed.


Lemma seg_fun_lub_weak_eq_iff :
  forall {n} (f:seg_fun n nat) s,
    0 < n ->
    (seg_fun_lub_weak f = s <->
     (forall i (pfi:i < n),
       f i pfi < S s)
     /\
     (forall b,
        (forall i (pfi:i < n),
           f i pfi < b) -> 
        S s <= b)).
  intros n f s h1.
  pose proof (seg_fun_lub_compat f h1) as h2.
  unfold seg_fun_lub in h2.
  split; intro h3; subst; auto with arith.
  destruct h2 as [h2a h2b], h3 as [h3a h3b].
  specialize (h3b (S (seg_fun_lub_weak f)) h2a).
  specialize (h2b (S s) h3a).
  omega.
Qed.


(*Adds the pair [m |-> y] to an [f:seg_fun], and shifts the old 
[m -> f m] to [S m -> f m], and likewise for function pairs above.*)
Fixpoint add_seg_fun {U n} m (y:U) :
  seg_fun n U -> seg_fun (S n) U :=
  match n with
    | 0 => fun _ _ _ => y
    | S n' =>
      match m with
        | 0 => 
          fun f i =>
            match i with
              | 0 =>
                fun _ => y
              | S i' =>
                fun pfi =>
                  let pfi := lt_S_n _ _ pfi in
                  f i' pfi
            end
        | S m' =>
          fun f i =>
            let f' := seg_fun_to_S f in
            match i with
              | 0 =>
                fun _ => 
                  f 0 (lt_0_Sn _)
              | S i' =>
                fun pfi =>
                  let pfi' := lt_S_n _ _ pfi in
                  add_seg_fun m' y f' i' pfi'
            end
      end
  end.
    
Lemma add_seg_fun_lt_compat :
  forall {U n} m y (f:seg_fun n U) i
         (pfi:  i < n)
         (pfi': i < S n), 
    i < m ->
    add_seg_fun m y f i pfi' = f i pfi.
  intros U n.
  induction n as [|n ih]; simpl.
  intros; omega.
  intro m; destruct m as [|m].
  intros; omega.
  intros y f i; destruct i as [|i].
  intros; f_equal; apply proof_irrelevance. 
  intros hi hi' h1.
  specialize (ih m y (seg_fun_to_S f) i (lt_S_n _ _ hi) (lt_S_n _ _ hi')).
  hfirst ih; auto with arith.
  specialize (ih hf).
  rewrite ih.
  unfold seg_fun_to_S.
  f_equal; apply proof_irrelevance.
Qed.


Lemma add_seg_fun_compat_eq :
  forall {U n} m y (f:seg_fun n U)
         (pfm:m < S n),
    m < n ->
    add_seg_fun m y f m pfm = y. 
  intros U n.
  induction n as [|n ih]; simpl; auto.
  intro m.
  destruct m as [|m]; auto.
  intros y f hm h1.
  rewrite ih; auto with arith.
Qed.

Lemma add_seg_fun_compat_gt :
  forall {U n} m y (f:seg_fun n U) i
         (pfi: i < S n)
         (pfi': pred i < n),
    i > m ->
    add_seg_fun m y f i pfi = f (pred i) pfi'.
  intros U n.
  induction n as [|n ih]; simpl.
  intros; omega.
  intro m; destruct m as [|m]. 
  intros y f i; destruct i as [|i]. 
  intros; omega.
  simpl; intros; f_equal; apply proof_irrelevance.
  intros y f i. 
  destruct i as [|i]. 
  simpl; intros; f_equal; apply proof_irrelevance.
  intros hi hi' h1.
  simpl.
  erewrite ih; auto with arith.
  unfold seg_fun_to_S.
  destruct i as [|i].
  omega.
  simpl.
  f_equal.
  apply proof_irrelevance.

  Grab Existential Variables.
  simpl in hi'.
  omega.
Qed.

                  
(*Removes one [pair-value] pair in a [seg_fun] at [m:nat] and shifts down 
the greater part of the bipartite partition.*)

Fixpoint remove_seg_fun {U n} m : seg_fun n U -> seg_fun (pred n) U :=
  match n return (seg_fun n U -> seg_fun (pred n) U ) with
    | 0 => fun _ _  pfi => False_rect _ (lt_n_0 _ pfi)
    | S n' =>
      fun f =>
        let f' := seg_fun_to_S f in
        let f_ := seg_fun_from_S f in
        match m with
          | 0 => f'
          | S m' =>
            fun i => 
              match i with
                | 0 =>
                  fun pfi => f_ 0 pfi
                | S i' =>
                  fun pfi => 
                    let pfi' := lt_S_pred _ _ pfi in
                    remove_seg_fun m' f' i' pfi'
              end
        end
  end.

Lemma remove_seg_fun_functional :
  forall {U n} m m' (f:seg_fun n U),
    m = m' ->
    remove_seg_fun m f = remove_seg_fun m' f.
  intros; subst; auto.
Qed.




Lemma remove_seg_fun_lt :
  forall {U n} (f:seg_fun n U) m i
         (pfi:i < pred n),
    i < m ->
    let pfi' := lt_pred' _ _ pfi in
    remove_seg_fun m f i pfi =
    f i pfi'.
  intros U n.
  induction n as [|n ih].
  simpl; intros; omega.
  simpl.
  intros f m.
  destruct m as [|m].
  intros; omega.
  intro i.
  destruct i as [|i].
  unfold seg_fun_from_S.
  intros; f_equal; apply proof_irrelevance.
  intros hi h1.
  apply lt_S_n in h1.
  specialize (ih (seg_fun_to_S f) m i (lt_S_pred _ _ hi) h1).
  rewrite ih.
  unfold seg_fun_to_S.
  f_equal.
  apply proof_irrelevance.
Qed.

Lemma remove_seg_fun_ge :
  forall {U n} (f:seg_fun n U) m i
         (pfi:i < pred n),
    i >= m ->
    let pfi' := lt_pred_S _ _ pfi in
    remove_seg_fun m f i pfi =
    f (S i) pfi'.
  intros U n.
  induction n as [|n ih].
  simpl; intros; omega.
  intros f m.
  destruct m as [|m].
  intros; simpl.
  unfold seg_fun_to_S; f_equal; apply proof_irrelevance.
  simpl.
  intro i.
  destruct i as [|i].
  intros; omega.
  intros hi h1.
  rewrite ih; auto with arith.
  unfold seg_fun_to_S.
  f_equal.
  apply proof_irrelevance.
Qed.


Lemma remove_seg_fun_ge_pred :
  forall {U n} m (f:seg_fun n U),
    m >= pred n ->
    remove_seg_fun m f = remove_seg_fun (pred n) f.
  intros U n. 
  induction n as [|n ih]; auto.
  intro m.
  destruct m as [|m].
  intros ? h1.
  apply le0 in h1; subst; auto.
  simpl in h1; subst; simpl; auto.
  intros f h1.
  simpl in h1.
  rewrite (remove_seg_fun_functional _ _ _ (eq_refl n)).
  simpl.
  destruct n as [|n].
  apply functional_extensionality_dep.
  intro i.
  destruct i as [|i]; apply functional_extensionality_dep; intro; omega.
  apply le_S_n in h1.
  apply functional_extensionality_dep.
  intro i.
  destruct i as [|i]; auto.
  apply functional_extensionality_dep.
  intro h2.
  pose proof (lt_S_n _ _ h2) as h3.
  specialize (ih _ (seg_fun_to_S f) h1). 
  pose proof (equal_f_dep _ _ ih _ (lt_S_pred _ _ h2)) as h4; auto.
Qed.

Lemma remove_seg_fun_to_S :
  forall {U n} m (f:seg_fun (S n) U),
    0 < m ->
    forall (pfn:0 < n),
        remove_seg_fun (S m) f =
        seg_fun_transfer (eq_sym (S_pred _ _ pfn))
                         (seg_fun_cons (f 0 (lt_0_Sn _))
                                       (remove_seg_fun m (seg_fun_to_S f))).
                                       
  intros; simpl.
  apply functional_extensionality_dep.
  intro i.
  destruct i as [|i].
  apply functional_extensionality.
  intro h1.
  rewrite seg_fun_transfer_compat.
  simpl.
  unfold seg_fun_from_S; f_equal; apply proof_irrelevance.
  apply functional_extensionality_dep.
  intro h1.
  pose proof (lt_eq_trans _ _ _ h1 (S_pred _ _ pfn)) as h2.
  apply lt_S_n in h2.
  rewrite seg_fun_transfer_compat.
  destruct (lt_le_dec i m) as [h3 | h3]. 
  rewrite remove_seg_fun_lt; auto.
  simpl.
  rewrite remove_seg_fun_lt; auto.
  repeat f_equal; apply proof_irrelevance.
  rewrite remove_seg_fun_ge; auto.
  simpl.
  rewrite remove_seg_fun_ge; auto.
  repeat f_equal; apply proof_irrelevance.
Qed.

Lemma remove_seg_fun_eq :
  forall {U n} m (f:seg_fun n U),
    0 < pred m ->
    forall (pfn:0 < n)
           (pfp:0 < pred n),
    remove_seg_fun m f =
    seg_fun_transfer (eq_sym (S_pred _ _ pfp))
                     (seg_fun_cons
                        (f 0 pfn)
                        (remove_seg_fun (pred m)
                                        (seg_fun_to_S
                                           (seg_fun_transfer
                                              (S_pred _ _ pfn) f)))).
  intros U n.
  destruct n as [|n].
  simpl.
  intros; omega.
  intros m f h1 h2 h3. 
  destruct m as [|m].
  contradict (lt_irrefl _ h1).
  erewrite remove_seg_fun_to_S; auto.
  repeat f_equal.
  apply proof_irrelevance.
  apply functional_extensionality_dep.
  intro.
  apply functional_extensionality.
  intro.
  rewrite seg_fun_transfer_compat.
  f_equal.
  apply proof_irrelevance.
Qed.


Lemma remove_seg_fun_to_S_swap0 :
  forall {U n} (f:seg_fun (S n) U)
         (pf:0 < n),
    remove_seg_fun 0 (seg_fun_to_S f) =
    seg_fun_to_S (seg_fun_transfer (S_pred _ _ pf) (remove_seg_fun 0 f)).
  intros; apply functional_extensionality_dep; intro; apply functional_extensionality; intro.
  rewrite remove_seg_fun_ge; auto with arith.
  unfold seg_fun_to_S.
  rewrite seg_fun_transfer_compat.
  rewrite remove_seg_fun_ge; auto with arith.
  f_equal; apply proof_irrelevance.
Qed.
  


  
Lemma remove_seg_fun_to_S_swap_pos :
  forall {U n} (f:seg_fun (S n) U) r
         (pf:0 < n),
    0 < r ->
    remove_seg_fun r (seg_fun_to_S f) =
    seg_fun_to_S (seg_fun_transfer (S_pred _ _ pf)
                                   (remove_seg_fun (S r) f)). 
  intros U n.
  destruct n as [|n].
  intros; omega. 
  intros f r h0 hr. 
  apply functional_extensionality_dep.
  intro i. 
  apply functional_extensionality.
  intro hi.
  destruct (lt_le_dec i r) as [h2|h2]; subst.
  rewrite remove_seg_fun_lt; auto with arith.
  unfold seg_fun_to_S.
  rewrite seg_fun_transfer_compat.
  rewrite remove_seg_fun_lt; auto with arith.
  f_equal; apply proof_irrelevance.
  unfold seg_fun_to_S; rewrite  seg_fun_transfer_compat.
  rewrite remove_seg_fun_ge; auto with arith.
  rewrite remove_seg_fun_ge; auto with arith.
  f_equal; apply proof_irrelevance.
Qed.


Lemma seg_fun_to_S_remove1 :
  forall {U n}
         (f:seg_fun (S (S n)) U),
    seg_fun_to_S (remove_seg_fun 1 f) =
    seg_fun_to_S (seg_fun_to_S f).
  intros; apply functional_extensionality_dep; intro i;
  apply functional_extensionality; intro hi;
  unfold seg_fun_to_S; rewrite remove_seg_fun_ge; auto with arith.
  f_equal; apply proof_irrelevance.
Qed.

Lemma seg_fun_remove0_to_S  : 
  forall {U n}
         (f:seg_fun (S (S n)) U),
   remove_seg_fun 0 (seg_fun_to_S f) = 
   seg_fun_to_pred (seg_fun_to_S f).
  intros; apply functional_extensionality_dep;
  intro i; apply functional_extensionality; intro hi;
  rewrite remove_seg_fun_ge; auto with arith;
  unfold seg_fun_to_pred, seg_fun_to_S; f_equal;
  try apply proof_irrelevance.
Qed.



Lemma seg_fun_from_S_remove :
  forall {U n}
         (f:seg_fun (S (S n)) U)
         r,
    seg_fun_from_S (remove_seg_fun r f) =
    remove_seg_fun r (seg_fun_from_S f).
  intros; apply functional_extensionality_dep; intro i;
  apply functional_extensionality; intro hi;
  rewrite seg_fun_from_S_compat.
  destruct (lt_le_dec i r) as [h2|h2]; subst.
  do 2 (rewrite remove_seg_fun_lt; auto).
  rewrite seg_fun_from_S_compat; f_equal; apply proof_irrelevance.
  do 2 (rewrite remove_seg_fun_ge; auto).
  rewrite seg_fun_from_S_compat; f_equal; apply proof_irrelevance.
Qed.


  

Lemma seg_fun_endo_transfer :
  forall {n n'} (f:seg_fun n nat) r
         (pfe:n = n'),
    seg_fun_endo f r ->
    seg_fun_endo (seg_fun_transfer pfe f) r.
  intros; subst; rewrite seg_fun_transfer_eq_refl; auto.
Qed.

Lemma inj_seg_fun_endo_transfer :
  forall {n n'} (f:seg_fun n nat) r
         (pfe:n = n'),
    inj_seg_fun_endo f r ->
    inj_seg_fun_endo (seg_fun_transfer pfe f) r.
  intros; subst; rewrite seg_fun_transfer_eq_refl; auto.
Qed.

Lemma surj_seg_fun_endo_transfer :
  forall {n n'} (f:seg_fun n nat) r
         (pfe:n = n'),
    surj_seg_fun_endo f r ->
    surj_seg_fun_endo (seg_fun_transfer pfe f) r.
  intros; subst; rewrite seg_fun_transfer_eq_refl; auto.
Qed.

Lemma seg_fun_endo_transfer' :
  forall {n n'} (f:seg_fun n nat) r
         (pfe:n = n'),
    seg_fun_endo (seg_fun_transfer pfe f) r ->
    seg_fun_endo f r.
  intros ? ? ? ? ? h1; subst. rewrite seg_fun_transfer_eq_refl in h1; auto.
Qed.

Lemma inj_seg_fun_endo_transfer' :
  forall {n n'} (f:seg_fun n nat) r
         (pfe:n = n'),
    inj_seg_fun_endo (seg_fun_transfer pfe f) r ->
    inj_seg_fun_endo f r.
  intros ? ? ? ? ? h1; subst. rewrite seg_fun_transfer_eq_refl in h1; auto.
Qed.


Lemma surj_seg_fun_endo_transfer' :
  forall {n n'} (f:seg_fun n nat) r
         (pfe:n = n'),
    surj_seg_fun_endo (seg_fun_transfer pfe f) r ->
    surj_seg_fun_endo f r.
  intros ? ? ? ? ? h1; subst. rewrite seg_fun_transfer_eq_refl in h1; auto.
Qed.


    


Lemma remove_seg_fun0 :
  forall {U n}
         (f:seg_fun (S n) U),
    remove_seg_fun 0 f = seg_fun_to_S f.
  intros ? n; destruct n; simpl; auto.
Qed.

Lemma remove_seg_fun0':
  forall {U n}
         (f:seg_fun n U),
    remove_seg_fun 0 f = seg_fun_to_pred f.
  intros ? n; destruct n; simpl; auto.
Qed.


Lemma seg_fun_to_S_eq :
  forall {U n} (f:seg_fun (S n) U),
    seg_fun_to_S f = remove_seg_fun 0 f.
  intros ? n; destruct n; simpl; auto.
Qed.

Lemma seg_fun_from_S_eq :
  forall {U n} (f:seg_fun (S n) U),
    seg_fun_from_S f = remove_seg_fun n f.
  intros U n.
  destruct n as [|n].
  intro f.
  apply functional_extensionality_dep.
  intro i.
  apply functional_extensionality.
  intro; omega.
  intro f.
  apply functional_extensionality_dep.
  intro i.
  destruct i as [|i].
  apply functional_extensionality.
  intro; auto.
  apply functional_extensionality.
  intro hi.
  rewrite remove_seg_fun_lt; auto.
  unfold seg_fun_from_S.
  f_equal; apply proof_irrelevance.
Qed.


Lemma seg_fun_to_S_remove_seg_fun1 :
  forall {n} (f:seg_fun (S (S n)) nat),
    seg_fun_to_S (remove_seg_fun 1 f) =
    seg_fun_to_S (seg_fun_to_S f).
  intros n f; apply functional_extensionality_dep;
  intro i; apply functional_extensionality;
  intro hi;
  unfold remove_seg_fun, seg_fun_to_S;
  f_equal; try apply proof_irrelevance.
Qed.





Lemma inv_seg_fun_eq_same_iff :
  forall {U n} (pfdu:type_eq_dec U)
         (f:seg_fun n U) i (pfi:i < n),
    (inv_seg_fun pfdu (f i pfi) f = i <->
     forall j (pfj:j < i) (pfj':j < n),
       f j pfj' <> f i pfi).
  intros U n.
  induction n as [|n ih]; simpl.
  intro hdu; intros f i hi.
  contradict (lt_n_O _ hi).
  intros hdu f i hi.
  destruct i as [|i]. 
  erewrite f_equal_dep.
  rewrite eq_dec_same at 1; auto.
  intuition.
  reflexivity.
  lr_if_goal; fold hlr; destruct hlr as [hl | hr].
  split; intro h1; try discriminate.
  specialize (h1 0 (lt_0_Sn _) (lt_0_Sn _)).
  contradiction.
  pose proof (lt_S_n _ _ hi) as h1.
  specialize (ih hdu (seg_fun_to_S f) i h1).
  unfold seg_fun_to_S in ih at 1.
  split.
  intros h3 j hj hj'.
  apply S_inj in h3.
  destruct j as [|j].
  erewrite f_equal_dep.
  apply hr.
  reflexivity.
  erewrite f_equal_dep in h3.
  rewrite ih in h3 at 1.
  unfold seg_fun_to_S in h3.
  erewrite f_equal_dep.
  apply neq_sym. 
  erewrite f_equal_dep.
  apply neq_sym.
  apply h3.
  apply lt_S_n.
  apply hj.
  reflexivity.
  reflexivity.
  reflexivity.
  intros h2.
  f_equal.
  erewrite f_equal_dep.
  rewrite ih at 1.
  intros j h3 h4.
  unfold seg_fun_to_S.
  specialize (h2 (S j) (lt_n_S _ _ h3) (lt_n_S _ _ h4)).
  contradict h2.
  rewrite h2.
  f_equal; apply proof_irrelevance.
  reflexivity.

  Grab Existential Variables.
  apply lt_S_n in hj'; auto.
Qed.

Lemma seg_fun_endo_to_S' :
  forall {n} (f:seg_fun (S n) nat) r s,
    r <= s ->
    f 0 (lt_0_Sn _) < s ->
    seg_fun_endo (seg_fun_to_S f) r ->
    seg_fun_endo f s.
  intros n f r s hr h1 h2.
  inversion h2 as [? ? h3 h4]; subst.
  constructor.
  intros i hi.
  destruct i as [|i].
  erewrite f_equal_dep.
  apply h1.
  reflexivity.
  specialize (h3 _ (lt_S_n _ _ hi)).
  unfold seg_fun_to_S in h3.
  eapply lt_le_trans.
  erewrite f_equal_dep.
  apply h3.
  reflexivity.
  assumption.
Qed.

Lemma seg_fun_endo_from_S' :
  forall {n} (f:seg_fun (S n) nat) r s,
    r <= s ->
    f n (lt_n_Sn _) < s ->
    seg_fun_endo (seg_fun_from_S f) r ->
    seg_fun_endo f s. 
  intros n f r s hr h1 h2.
  inversion h2 as [? ? h3 h4]; subst.
  constructor.
  intros i hi.
  pose proof (lt_dec _ _ hi) as h4.
  simpl in h4.
  destruct h4 as [h4|]; subst; auto.
  specialize (h3 i h4).
  unfold seg_fun_from_S in h3.
  eapply lt_le_trans.
  erewrite f_equal_dep.
  apply h3.
  reflexivity.
  assumption.
  erewrite f_equal_dep.
  apply h1.
  reflexivity.
Qed.


Lemma seg_fun_endo_to_S_bound :
  forall {n} b (f:seg_fun (S n) nat),
    seg_fun_endo f b ->
    ~in_seg_fun (pred b) (seg_fun_to_S f) ->
    seg_fun_endo (seg_fun_to_S f) (pred b).
  intros n b f h1 h2.
  destruct h1 as [? ? h1].
  constructor.
  intros i hi.
  specialize (h1 (S i) (lt_n_S _ _ hi)).
  apply lt_dec in h1.
  destruct h1 as [h1|h1].
  unfold seg_fun_to_S; auto.
  contradict h2.
  rewrite <- h1.
  erewrite f_equal_dep.
  erewrite seg_fun_to_S_compat'.
  apply in_seg_fun_compat.
  reflexivity.

  Grab Existential Variables.
  auto with arith.
Qed.




Lemma shift_seg_fun_endo :
  forall n k,
    seg_fun_endo (shift_seg_fun n k) n.
  intros n k; constructor; intros; apply lt_shift_seg_fun.
Qed.

Lemma inj_seg_fun_endo_shift :
  forall n k,
    inj_seg_fun_endo (shift_seg_fun n k) n.
  intros; constructor; try apply shift_seg_fun_endo;
  apply inj_shift_seg_fun.
Qed.

Lemma seg_fun_endo_right_shift :
  forall n,
    seg_fun_endo (right_shift_seg_fun n) n.
  intros; apply shift_seg_fun_endo.
Qed.

Lemma seg_fun_endo_left_shift :
  forall n,
    seg_fun_endo (left_shift_seg_fun n) n.
  intros; apply shift_seg_fun_endo.
Qed.



Lemma add_seg_to_identify_endo :
  forall n,
    0 < n ->
    seg_fun_endo (add_seg_to_identify n) n.
  intros; constructor; auto; intros.
  apply lt_add_seg_to_identify; auto.
Qed.

Lemma tl_seg_for_fictitious_endo :
  forall n,
    seg_fun_endo (tl_seg_for_fictitious n) n.
  intro n.
  constructor.
  apply lt_tl_seg_for_fictitious.
Qed.



Lemma seg_fun_endo_remove :
  forall {n i m} (f:seg_fun n nat),
    i < n ->
    seg_fun_endo f m->
    seg_fun_endo (remove_seg_fun i f) m.
  intros n i m f h1 h2. 
  revert h1. revert i.
  induction h2 as [r f ih].
  intros i h1.
  econstructor.
  intros j hj. 
  destruct (le_lt_dec i j) as [h2 | h2].
  pose proof (remove_seg_fun_ge f i j hj h2) as h3.
  simpl in h3.
  rewrite h3; auto.
  pose proof (remove_seg_fun_lt f i j hj h2) as h3.
  simpl in h3.
  rewrite h3; auto.
Qed.

Lemma seg_fun_endo_remove' :
  forall {n r} (f:seg_fun n nat)
         b c (pfr:r < n) ,
    b <= c ->
    f r pfr < c ->
    seg_fun_endo (remove_seg_fun r f) b ->
    seg_fun_endo f c.
  intros n r f b c hb hc h1 h2.
  inversion h2 as [? ? h3 h4]; subst.
  destruct (zerop (pred n)) as [h4|h4].
  apply pred0_dec in h4.
  destruct h4 as [h4 | h4].
  contradict (lt_n_0 _ (lt_eq_trans _ _ _ hb h4)).
  pose proof (lt_eq_trans _ _ _ hb h4) as h5.
  apply lt1 in h5.
  revert h1. revert hb.
  subst.
  intros h0 h1.
  constructor.
  intros i hi.
  pose proof (lt1 _ hi); subst.
  erewrite f_equal_dep.
  apply h1.
  reflexivity.
  constructor.
  intros i hi.
  destruct (zerop r) as [|h5]; subst.
  destruct i as [|i].
  erewrite f_equal_dep.
  apply h1.
  reflexivity.
  specialize (h3 i (lt_S_pred _ _ hi)).
  rewrite remove_seg_fun_ge in h3.
  eapply lt_le_trans.
  erewrite f_equal_dep.
  apply h3.
  reflexivity.
  assumption.
  auto with arith.
  pose proof (lt_dec _ _ hi) as h6. 
  destruct (lt_le_dec i r) as [h7|h7]; subst.
  destruct h6 as [h6|]; subst. 
  specialize (h3 i h6).
  rewrite remove_seg_fun_lt in h3; auto.
  eapply lt_le_trans.
  erewrite f_equal_dep.
  apply h3.
  reflexivity.
  assumption.
  omega. 
  destruct h6 as [h6|]; subst.
  specialize (h3 (pred i)).
  hfirst h3.
  omega.
  specialize (h3 hf).
  apply le_dec' in h7.
  destruct h7 as [h7|]; subst.
  rewrite remove_seg_fun_ge in h3; auto.
  revert h3.
  generalize (lt_pred_S _ _ hf).
  assert (h8:0 < i).  omega.
  rewrite <- (S_pred _ _ h8).
  intros h9 h10.
  erewrite f_equal_dep.
  eapply lt_le_trans.
  apply h10.
  assumption.
  reflexivity.
  erewrite f_equal_dep.
  apply h1.
  reflexivity.
  apply le_dec' in h7.
  destruct h7 as [h7|]; subst.
  assert (h8:pred (pred n) < pred n). omega.
  specialize (h3 (pred (pred n)) h8).
  rewrite remove_seg_fun_ge in h3; auto.
  revert h3.
  generalize (lt_pred_S _ _ h8).
  rewrite <- (S_pred _ _ h4).
  intros h9 h10.
  eapply lt_le_trans.
  erewrite f_equal_dep.
  apply h10.
  reflexivity.
  assumption.
  erewrite f_equal_dep.
  apply h1.
  reflexivity.
Qed.


Lemma inj_seg_fun_endo_remove :
  forall {n i m} (f:seg_fun n nat),
    i < n ->
    inj_seg_fun_endo f m->
    inj_seg_fun_endo (remove_seg_fun i f) m.
  intros n i m f h1 h2.
  constructor.
  apply seg_fun_endo_remove; auto.
  inversion h2; auto.
  intros r s hr hs h3.
  inversion h2; subst.
  destruct (lt_le_dec r i) as [h4 | h4]. 
  rewrite (remove_seg_fun_lt f i r hr h4) in h3. 
  destruct (lt_le_dec s i) as [h5 | h5].
  rewrite (remove_seg_fun_lt f i s hs h5) in h3.
  apply H0 in h3; auto.
  rewrite (remove_seg_fun_ge f i s hs h5) in h3.
  apply H0 in h3; subst.
  omega.
  rewrite (remove_seg_fun_ge f i r hr h4) in h3.
  destruct (lt_le_dec s i) as [h5 | h5].
  rewrite (remove_seg_fun_lt f i s hs h5) in h3.
  apply H0 in h3; subst; omega.
  rewrite (remove_seg_fun_ge f i s hs h5) in h3.
  apply H0 in h3.
  apply S_inj in h3; auto.
Qed.

Lemma inj_seg_fun_endo_to_S :
  forall {n} (f:seg_fun (S n) nat) r,
    inj_seg_fun_endo f r->
    inj_seg_fun_endo (seg_fun_to_S f) r.
  intros; rewrite seg_fun_to_S_eq.
  eapply inj_seg_fun_endo_remove in H.
  apply H.
  auto with arith.
Qed.

Lemma inj_seg_fun_endo_from_S :
  forall {n} (f:seg_fun (S n) nat) r,
    inj_seg_fun_endo f r->
    inj_seg_fun_endo (seg_fun_from_S f) r.
  intros; rewrite seg_fun_from_S_eq.
  eapply inj_seg_fun_endo_remove in H.
  apply H.
  auto with arith.
Qed.


  


Definition inj_seg_fun_lub {n}
           (f:seg_fun n nat) : Prop :=
  inj_seg_fun_endo f (seg_fun_lub f).

Definition inv_seg_fun_lub_weak {n}
           (f:seg_fun n nat) (pf0:0 < n) : nat :=
  inv_seg_fun nat_eq_dec (seg_fun_lub_weak f) f.


Definition remove_seg_fun_lub_weak {n} (f:seg_fun n nat) (pf0:0 < n) :
  seg_fun (pred n) nat :=
  remove_seg_fun (inv_seg_fun_lub_weak f pf0) f.

Lemma le_seg_fun_lub_weak_to_S :
  forall {n} (f:seg_fun (S n) nat),
    seg_fun_lub_weak (seg_fun_to_S f) <=
    seg_fun_lub_weak f.
  intro n.
  destruct n as [|n].
  simpl; auto with arith.
  intro f.
  simpl.
  destruct (gt_ge_dec (seg_fun_to_S f 0 (lt_0_Sn n))
                      (seg_fun_lub_weak (seg_fun_to_S (seg_fun_to_S f)))) as [h1 | h1].
  apply gt_not_ge in h1.
  rewrite <- ge_nat_compat in h1.
  rewrite <- eq_false_iff in h1.
  rewrite h1.
  destruct (gt_ge_dec (f 0 (lt_0_Sn (S n)))
                      (seg_fun_to_S f 0 (lt_0_Sn n))) as [h2 | h2].
  pose proof h2 as h2'.
  apply gt_not_ge in h2.
  rewrite <- ge_nat_compat in h2.
  rewrite <- eq_false_iff in h2.
  rewrite h2.
  omega.
  rewrite <- ge_nat_compat in h2.
  rewrite h2; auto.
  rewrite <- ge_nat_compat in h1.
  rewrite h1.
  destruct (gt_ge_dec (f 0 (lt_0_Sn (S n)))
                      (seg_fun_lub_weak (seg_fun_to_S (seg_fun_to_S f)))) as [h2 | h2].
  pose proof h2 as h2'.
  apply gt_not_ge in h2.
  rewrite <- ge_nat_compat in h2.
  rewrite <- eq_false_iff in h2.
  rewrite h2.
  omega.
  rewrite <- ge_nat_compat in h2.
  rewrite h2; auto.
Qed.



Lemma lt_seg_fun_lub_to_S_dec :
  forall {n} (f:seg_fun (S n) nat)
         (pf:0 < S n),
    {f 0 pf = seg_fun_lub_weak f} +
    {seg_fun_lub_weak f = seg_fun_lub_weak (seg_fun_to_S f)}.
  intro n.
  destruct n as [|n].
  simpl.
  intros f h1.
  destruct (zerop (f 0 (lt_0_Sn 0))) as [h2 | h2].
  rewrite h2.
  right; auto.
  rewrite (S_pred _ _ h2).
  rewrite <- (S_pred _ _ h2).
  left.
  apply f_equal_dep; auto.
  intros f h1.
  simpl.
  destruct (gt_ge_dec (seg_fun_to_S f 0 (lt_0_Sn n))
                      (seg_fun_lub_weak (seg_fun_to_S (seg_fun_to_S f)))) as [h2 | h2].
  pose proof h2 as h2'.
  apply gt_not_ge in h2'.
  rewrite <-  ge_nat_compat in h2'.
  rewrite <- eq_false_iff in h2'.
  rewrite h2'.
  destruct (gt_ge_dec (f 0 (lt_0_Sn (S n)))
                      (seg_fun_to_S f 0 (lt_0_Sn n)))
    as [h3 | h3].
  pose proof h3 as h3'.
  apply gt_not_ge in h3'.
  rewrite <-  ge_nat_compat in h3'.
  rewrite <- eq_false_iff in h3'.
  rewrite h3'.
  left.
  apply f_equal_dep; auto.
  pose proof h3 as h3'. 
  rewrite <-  ge_nat_compat in h3'.
  rewrite h3'.
  right; auto.
  pose proof h2 as h2'.
  rewrite <- ge_nat_compat in h2'.
  rewrite h2'.
  destruct (gt_ge_dec (f 0 (lt_0_Sn _))
                      (seg_fun_lub_weak (seg_fun_to_S (seg_fun_to_S f)))) as [h3 | h3].
  pose proof h3 as h3'.
  apply gt_not_ge in h3'.
  rewrite <-  ge_nat_compat in h3'.
  rewrite <- eq_false_iff in h3'.
  rewrite h3'.
  left; f_equal; auto; apply proof_irrelevance.
  pose proof h3 as h3'.
  rewrite <- ge_nat_compat in h3'.
  rewrite h3'.
  right; auto.
Qed.

Section PreImage.

Definition pre_class {X S} (v:X->S) (s:S) : Ensemble X :=
  [x:X | v x = s].

Lemma in_pre_class :
  forall {X S} (v:X->S) x,
    Inn (pre_class v (v x)) x.
  intros; constructor; auto.
Qed.

Lemma pre_class_inj :
  forall {X S} (v:X->S) (s s':S),
    in_pre s v ->
    pre_class v s =
    pre_class v s' ->
    s = s'.
  intros X S v s s' h1 h2. 
  destruct h1.
  pose proof (in_pre_class v x) as h3.
  rewrite h2 in h3.
  destruct h3; auto.
Qed.
                              
Lemma in_pre_inhabited :
  forall {X S} (v:X->S) (s:S),
    in_pre s v ->
    Inhabited (pre_class v s).
  intros X S v s h1.
  destruct h1.
  apply Inhabited_intro with x.
  constructor; auto.
Qed.


Lemma inh_in_pre : 
  forall {X S} (v:X->S) (s:S),
    Inhabited (pre_class v s) ->
    in_pre s v.
  intros X S v s h1.
  inversion h1 as [x h2].
  destruct h2 as [h2]; subst.
  constructor.
Qed.


Inductive pre_classes {X S} (v:X->S) : Family X :=
| pre_classes_intro : forall s,
                        in_pre s v ->
                        Inn (pre_classes v)
                            (pre_class v s).

Lemma inh_ind_fam_impl :
  forall {X S} (v:X->S),
    inh_ind_var v ->
    Inhabited (pre_classes v).
  intros X S v h1. 
  destruct h1 as [s h1].
  apply Inhabited_intro with (pre_class v s).
  constructor; auto.
Qed.

Lemma in_pre_classes :
  forall {S X} (v:X->S) x,
    Inn (pre_classes v) (pre_class v (v x)).
  intros; constructor; constructor.
Qed.


Lemma in_pre_impl : 
  forall {X S}
         {v:X->S}
         {s:S},
    in_pre s v ->
    Inn (pre_classes v) (pre_class v s).
  intros ? ? ? ? h1; destruct h1; constructor; constructor.
Qed.


Inductive pre_app_aux {X S} (v:X->S) :
  forall A:Ensemble X, Inn (pre_classes v) A -> S -> Prop :=
  pre_app_aux_intro :
    forall s (pfip:in_pre s v),
      pre_app_aux v (pre_class v s) (pre_classes_intro v s pfip) s.


Lemma pre_app_ex :
  forall {X S} (v:X->S) A (pfi:Inn (pre_classes v) A),
  exists! s:S,
    pre_app_aux v A pfi s /\
    A = pre_class v s.
  intros X S v A h1.
  inversion h1; subst.
  exists s.
  red; split.
  split; auto.
  pose proof (proof_irrelevance
                _ h1
                (pre_classes_intro v s H));
    subst.
  constructor.
  intros s' h2. 
  destruct h2 as [h2 h3].
  apply pre_class_inj in h3; auto.
Qed.


Definition pre_app
           {X S}
           (v:X->S)
           A (pfi:Inn (pre_classes v) A) :
  S :=
  proj1_sig (constructive_definite_description
               _
               (pre_app_ex v A pfi)).


Lemma pre_app_compat :
  forall {X S} (v:X->S)
         A (pfi:Inn (pre_classes v) A),
    let s := pre_app v A pfi in
    pre_app_aux v A pfi s /\
    A = pre_class v s.
  unfold pre_app; intros; destruct constructive_definite_description; auto.
Qed.


Inductive pre_app_inh_aux 
          {S X}
          {U:S->Type}
          (v:X->S)
          (a:forall x:X, U (v x)) :
  forall s (pfip:in_pre_aux v s), U s -> Prop :=
  pre_app_inh_aux_intro : forall x,
                            pre_app_inh_aux v a (v x)
                                            (in_pre_intro v x)
                                            (a x).

End PreImage.

Section SetSortedSegFun.  


(*This is a recursive predicate that has some advantages 
and disadvandages over the simpler [seg_fun]. 
[seg_fun] is great when you want simplicity 
and power, and want easy syntax, 
but bad if you do a lot of parameterizations with 
indices or inequality proofs, cause then you get a semantically
cumbersome syntax.  Those invariably necessitate 
[transfer] functions, which in general are a nuisance, 
and can be obviated if the terms in the defining equality are 
in a type in [Set].  Then we can compute the predicates in [bool] 
(rather than have them be in [Prop]).  
So, using an inductive predicate will be an important technique
to handle the [destructing] of [sseg_funs] ([Set]-seg_funs].
When we destruct terms, the process should be as painless as 
pattern-matching.*)
(*The trade-off of using this over [seg_fun] is less 
computational and metaphysical purity -- which is why 
this section only indicates how this would be gone about.*)

(*[sseg_fun offset n f offset] assures us that f has only 
[n] intended values (from 0 to [pred n]); the lingering values 
in the function are the same as the last value (at [pred n]).  
It also ensures us in the 3rd constructor that it was constructed 
from a previous function, with the adding of a value at [offset], 
and [n] shifted up once.*)
  
  Inductive sseg_fun_aux {U} : (nat->U) -> nat -> nat -> Prop :=
    sseg_fun0 : forall (u:U),
                  sseg_fun_aux (fun _ => u) 0 0
| sseg_fun1 : forall (u:U),
                sseg_fun_aux (fun _ => u) 1 0
| sseg_fun_S : forall (n offset offset':nat) (f:nat->U) (u:U),
                 sseg_fun_aux f n offset ->
                 sseg_fun_aux 
                   (fun i =>
                      if (lt_nat i (S n)) then
                        if (lt_nat i offset') then 
                          f i
                        else
                          if (beq_nat i offset') then
                            u
                          else
                            f (S i)
                      else
                        f n)
                   (S n)
                   (min offset' n).

  Inductive sseg_fun {U} (f:nat->U) n : Prop :=
    sseg_fun_intro : forall offset,
                       sseg_fun_aux f n offset ->
                       sseg_fun f n.
  
  Inductive sseg_fun_remove_aux {U} :
    forall (f:nat->U) (n:nat) (offset:nat),
      sseg_fun_aux f (S n) offset ->
      forall (f':nat->U), sseg_fun f' n -> Prop :=
    seg_fun_remove_aux_intro :
      forall f n old_offset offset
           (pfss:sseg_fun_aux f n old_offset)
           u,
      let pfss' := sseg_fun_S n old_offset offset f u pfss in
      let pfss_ := sseg_fun_intro f n _ pfss in
      sseg_fun_remove_aux
        _ n (min offset n) pfss' f pfss_.
  
End SetSortedSegFun.

Definition l_to_fun {T} (l:list T) (pfl:l <> nil) : nat -> T :=
  fun i => nth i l (lst l pfl).

Inductive segf {T} : (nat -> T) -> Prop :=
| segf_intro : forall (l:list T) (pfl:l <> nil),
                 segf (l_to_fun l pfl).

Lemma in_seg_fun_list_iff :
  forall {U n} (f:seg_fun n U) y,
    in_seg_fun y f <->
    In y (seg_fun_to_list f).
  intros; split; intro h1.
  rewrite in_seg_fun_iff in h1.
  destruct h1 as [? [h1 ?]]; subst.
  apply in_map_seg.
  rewrite in_map_seg_iff in h1.
  destruct h1 as [? [? ?]]; subst.
  apply in_seg_fun_compat.
Qed.

Lemma inj_seg_fun_endo_to_S' :
  forall {n} (f:seg_fun (S n) nat) r s,
    r <= s ->
    f 0 (lt_0_Sn _) < s ->
    inj_seg_fun_endo (seg_fun_to_S f) r ->
    ~in_seg_fun (f 0 (lt_0_Sn _)) (seg_fun_to_S f) ->
    inj_seg_fun_endo f s.
  intros n f r s hr h1 h2 h3.
  inversion h2 as [? ? h4 h5]; subst.
  inversion h4 as [? ? h6 h7]; subst.
  constructor.
  eapply seg_fun_endo_to_S'.
  apply hr.
  assumption.
  assumption.
  intros i j hi hj h7.
  destruct i as [|i], j as [|j]; subst; auto.
  contradict h3.
  erewrite f_equal_dep.
  rewrite h7 at 1.
  erewrite f_equal_dep.
  erewrite seg_fun_to_S_compat' at 1.
  apply in_seg_fun_compat.
  reflexivity.
  reflexivity.
  contradict h3.
  erewrite f_equal_dep.
  rewrite <- h7 at 1.
  erewrite f_equal_dep.
  erewrite seg_fun_to_S_compat' at 1.
  apply in_seg_fun_compat.
  reflexivity.
  reflexivity. 
  specialize (h5 i j (lt_S_n _ _ hi) (lt_S_n _ _ hj)).
  unfold seg_fun_to_S in h5.
  hfirst h5.
  erewrite f_equal_dep.
  rewrite h7 at 1.
  f_equal; apply proof_irrelevance.
  reflexivity.
  specialize (h5 hf); auto with arith.

  Grab Existential Variables.
  auto with arith.
  auto with arith.
Qed.

Lemma inj_seg_fun_endo_to_S_bound :
  forall {n} b (f:seg_fun (S n) nat),
    inj_seg_fun_endo f b ->
    ~in_seg_fun (pred b) (seg_fun_to_S f) ->
    inj_seg_fun_endo (seg_fun_to_S f) (pred b).
  intros n b f h1 h2.
  inversion h1 as [? ? h3 h4]; subst.
  constructor.
  apply seg_fun_endo_to_S_bound; auto.
  intros i j hi hj h5.
  specialize (h4 (S i) (S j) (lt_n_S _ _ hi) (lt_n_S _ _ hj) h5).
  apply S_inj in h4; auto.
Qed.


Lemma inj_seg_fun_endo_remove' :
  forall {n} (f:seg_fun n nat) r s m
         (pfm: m < n),
    r <= s ->
    f m pfm < s ->
    inj_seg_fun_endo (remove_seg_fun m f) r ->
    ~in_seg_fun (f m pfm) (remove_seg_fun m f) ->
    inj_seg_fun_endo f s. 
  intro n.
  induction n as [|n ih].
  intros; omega.
  intros f r s m hm.
  destruct m as [|m].
  intros h1 h2 h3 h4.
  inversion h3 as [? ? h5 h6]; subst.
  constructor.
  eapply seg_fun_endo_to_S'.
  apply h1.
  erewrite f_equal_dep.
  apply h2.
  reflexivity.
  assumption.
  intros i j hi hj h7.
  destruct i as [|i], j as [|j]; auto.
  contradict h4.
  erewrite f_equal_dep.
  rewrite h7 at 1.
  erewrite f_equal_dep.
  erewrite seg_fun_to_S_compat'.
  apply in_seg_fun_compat.
  reflexivity.
  reflexivity.
  contradict h4.
  erewrite f_equal_dep.
  rewrite <- h7 at 1.
  erewrite f_equal_dep.
  erewrite seg_fun_to_S_compat' at 1.
  apply in_seg_fun_compat.
  reflexivity.
  reflexivity. 
  specialize (h6 i j (lt_S_n _ _ hi) (lt_S_n _ _ hj)).
  f_equal.
  apply h6.
  simpl.
  unfold seg_fun_to_S.
  erewrite f_equal_dep.
  rewrite h7 at 1.
  f_equal; apply proof_irrelevance.
  reflexivity.
  pose proof (lt_S_n _ _ hm) as h0.
  intros hr h1 h2 h3.
  inversion h2 as [? ? h5 h6]; subst.
  inversion h5 as [? ? h7 h8]; subst.
  constructor.
  eapply seg_fun_endo_remove' in h5.
  apply h5.
  assumption.
  apply h1.
  intros i j hi hj hk. 
  destruct i as [|i], j as [|j]; auto.
  pose proof (lt_S_n _ _ hj) as h8. 
  destruct (lt_le_dec j m) as [h9 | h9]. 
  pose proof h6 as h6'.
  specialize (h6 0 (S j) (Olt _ _ h8)).
  pose proof (lt_dec _ _ hj) as h10.
  simpl in h10.
  destruct h10 as [h10|].
  revert h1.
  subst.
  intro h1.
  specialize (h6 h10).
  hfirst h6. 
  rewrite remove_seg_fun_lt; auto with arith. 
  rewrite remove_seg_fun_lt; auto.
  erewrite f_equal_dep.
  rewrite hk at 1.
  f_equal.
  apply proof_irrelevance.
  reflexivity.
  auto with arith.
  specialize (h6 hf); discriminate. 
  revert h1 hk; subst; intros h1 hk. 
  destruct (lt_le_dec j (S m)) as [h9' | h9']. 
  omega.
  specialize (h6' 0 j).
  hfirst h6'.
  simpl; auto with arith.
  specialize (h6' hf).
  hfirst h6'.
  simpl.
  auto with arith.
  specialize (h6' hf0).
  hfirst h6'.
  rewrite remove_seg_fun_lt; auto with arith.
  rewrite remove_seg_fun_ge; auto with arith.
  erewrite f_equal_dep.
  rewrite hk at 1. 
  apply f_equal; apply proof_irrelevance.
  reflexivity.
  specialize (h6' hf1).
  revert h1; subst; intro h1.
  omega.
  apply le_dec' in h9.
  destruct h9 as [h9 | h9].
  pose proof h6 as h6'.
  specialize (h6 0 j (Olt _ _ h8) (lt_S_n _ _ hj)).
  hfirst h6.
  rewrite remove_seg_fun_lt; auto with arith.
  rewrite remove_seg_fun_ge.
  erewrite f_equal_dep.
  rewrite hk at 1.
  f_equal; apply proof_irrelevance.
  reflexivity.
  destruct j as [|j]; subst.
  apply le0 in h9; subst.
  contradict h3.  
  rewrite in_seg_fun_iff.  
  exists 0.
  ex_goal.
  simpl; auto with arith.
  exists hex. 
  rewrite remove_seg_fun_lt; auto with arith.
  erewrite f_equal_dep.
  rewrite hk at 1.
  f_equal; apply proof_irrelevance.
  reflexivity.
  auto with arith.
  specialize (h6 hf); subst.
  contradict h3.
  rewrite in_seg_fun_iff.
  exists 0.
  ex_goal.
  simpl; auto with arith.
  exists hex.
  rewrite remove_seg_fun_lt; auto with arith.
  apply le0 in h9; subst.
  erewrite f_equal_dep.
  rewrite hk at 1.
  f_equal; apply proof_irrelevance.
  reflexivity.
  subst.
  contradict h3.
  erewrite f_equal_dep.
  rewrite <- hk at 1.
  rewrite in_seg_fun_iff.
  exists 0, (Olt _ _ h0).
  rewrite remove_seg_fun_lt; auto with arith.
  f_equal.
  apply proof_irrelevance.
  reflexivity.
  pose proof (lt_dec _ _ hi) as h8.
  simpl in h8.
  destruct h8 as [h8|]; revert h1 hk; subst; intros h1 hk.
  destruct (lt_le_dec i m) as [h10|h10].
  specialize (h6 (S i) 0 h8 (Olt _ _ h0)).
  hfirst h6.
  rewrite remove_seg_fun_lt; auto with arith.
  rewrite remove_seg_fun_lt; auto with arith.
  erewrite f_equal_dep.
  rewrite hk at 1.
  f_equal; apply proof_irrelevance.
  reflexivity.
  apply h6 in hf; discriminate.
  apply le_dec' in h10.
  destruct h10 as [h10|]; subst.
  specialize (h6 i 0 (lt_S_n _ _ hi) (Olt _ _ h0)). 
  hfirst h6.
  rewrite remove_seg_fun_ge.
  rewrite remove_seg_fun_lt; auto with arith.
  erewrite f_equal_dep.
  rewrite hk at 1.
  f_equal; apply proof_irrelevance.
  reflexivity.
  destruct i as [|i]. 
  apply le0 in h10; subst.
  contradict h3.
  erewrite f_equal_dep.
  rewrite hk at 1.
  rewrite in_seg_fun_iff.
  exists 0, h0.
  rewrite remove_seg_fun_lt; f_equal.
  apply proof_irrelevance.
  auto with arith.
  reflexivity.
  simpl in h10; auto with arith. 
  specialize (h6 hf).
  subst. 
  contradict h3.
  apply le0 in h10; subst.
  rewrite in_seg_fun_iff.
  exists 0, h0.
  simpl.
  rewrite seg_fun_from_S_compat.
  erewrite f_equal_dep.
  rewrite <- hk at 1.
  f_equal.
  apply proof_irrelevance.
  reflexivity.
  contradict h3.
  rewrite in_seg_fun_iff.
  exists 0, (Olt _ _ h0).
  rewrite remove_seg_fun_lt; auto with arith.
  erewrite f_equal_dep.
  rewrite <- hk at 1.
  f_equal; apply proof_irrelevance.
  reflexivity.
  red in h0.
  apply le_S_n in h0.
  apply le_dec' in h0.
  destruct h0 as [h0|]; subst.
  specialize (h6 i 0 (lt_n_Sn _) (lt_0_Sn _)).
  hfirst h6.
  rewrite remove_seg_fun_ge.
  rewrite remove_seg_fun_lt; auto with arith.
  erewrite f_equal_dep.
  rewrite hk at 1.
  f_equal; apply proof_irrelevance.
  reflexivity.
  destruct i as [|i].
  apply le0 in h0; subst.
  contradict h3.
  rewrite in_seg_fun_iff.
  exists 0, lt_0_1.
  rewrite remove_seg_fun_lt.
  erewrite f_equal_dep.
  rewrite <- hk at 1.
  f_equal; apply proof_irrelevance.
  reflexivity.
  auto with arith.
  auto with arith.
  specialize (h6 hf); revert hm h3 h1; subst; intros hm h3 h1.
  apply le0 in h0; subst.
  contradict h3.
  rewrite in_seg_fun_iff.
  exists 0, lt_0_1.
  rewrite remove_seg_fun_lt.
  erewrite f_equal_dep.
  rewrite <- hk at 1.
  f_equal; apply proof_irrelevance.
  reflexivity.
  auto with arith.
  contradict h3.
  rewrite in_seg_fun_iff.
  exists 0, (lt_0_Sn _).
  rewrite remove_seg_fun_lt; auto with arith.
  erewrite f_equal_dep.
  rewrite <- hk at 1.
  f_equal; apply proof_irrelevance.
  reflexivity.   
  specialize (ih (seg_fun_to_S f) r s m h0 hr).
  hfirst ih.
  unfold seg_fun_to_S.
  erewrite f_equal_dep.
  apply h1.
  reflexivity.
  specialize (ih hf).
  hfirst ih.
  destruct (zerop m) as [|h8]; subst.
  destruct n as [|n].
  omega.
  rewrite remove_seg_fun0.
  apply inj_seg_fun_endo_to_S in h2.
  rewrite seg_fun_to_S_remove_seg_fun1 in h2; auto.  
  erewrite remove_seg_fun_to_S_swap_pos.
  apply inj_seg_fun_endo_to_S.
  apply inj_seg_fun_endo_transfer.
  assumption.
  assumption. 
  specialize (ih hf0). 
  hfirst ih. 
  contradict h3.
  rewrite in_seg_fun_iff in h3.
  rewrite in_seg_fun_iff.
  destruct h3 as [k [h10 h11]].
  destruct (lt_le_dec k m) as [h12|h12]; subst.
  rewrite remove_seg_fun_lt in h11; auto.
  exists (S k).
  ex_goal.
  simpl; omega.
  exists hex.
  rewrite remove_seg_fun_lt; auto with arith.
  unfold seg_fun_to_S in h11.
  erewrite f_equal_dep.
  rewrite h11 at 1.
  f_equal; apply proof_irrelevance.
  reflexivity.
  rewrite remove_seg_fun_ge in h11; auto with arith.
  unfold seg_fun_to_S in h11.
  exists (S k).
  ex_goal.
  omega.
  exists hex.
  rewrite remove_seg_fun_ge; auto with arith.
  erewrite f_equal_dep.
  rewrite h11 at 1.
  f_equal; apply proof_irrelevance.
  reflexivity.
  specialize (ih hf1).
  inversion ih as [? ? h11 h12]; subst.
  specialize (h12 i j (lt_S_n _ _ hi) (lt_S_n _ _ hj)).
  hfirst h12.
  unfold seg_fun_to_S.
  erewrite f_equal_dep.
  rewrite hk at 1.
  f_equal; apply proof_irrelevance.
  reflexivity.
  specialize (h12 hf2); subst; auto.

  Grab Existential Variables.
  destruct (zerop n); revert hm h1 h3 hk; subst; auto; intros hm h1 h3 hk.
  omega.
  auto with arith.
  auto with arith.
Qed.


Lemma inv_seg_fun_lt :
  forall {U n} (pfdu:type_eq_dec U) (y:U) (f:seg_fun n U),
    in_seg_fun y f ->
    inv_seg_fun pfdu y f < n.
  intros U n. 
  induction n as [|n ih]; simpl; auto with arith.
  intros; contradiction.
  intros hdu y; simpl.
  intro f.
  destruct (hdu (f 0 (lt_0_Sn _)) y) as [|h2]; subst; auto with arith.
  intro h1.
  apply lt_n_S; auto.
  apply ih. 
  destruct h1; try contradiction; auto.
Qed.

Lemma inv_seg_fun_dec_lt :
  forall {U n} (pfdu:type_eq_dec U) (y:U) (f:seg_fun n U),
    in_seg_fun_dec y f ->
    inv_seg_fun pfdu y f < n.
  intros; apply inv_seg_fun_lt.
  apply in_seg_fun_dec_impl; auto.
Qed.


Lemma inv_seg_fun_eq_bound_iff :
  forall {U n} (pfdu:type_eq_dec U)
         (f:seg_fun n U) y,
    inv_seg_fun pfdu y f = n <->
    ~in_seg_fun y f.
  intros U n.
  induction n as [|n ih]; simpl.
  intuition.
  intros hdu f y.
  destruct (hdu (f 0 (lt_0_Sn n)) y) as [h1 | h1].
  intuition.
  split; intro h2.
  apply S_inj in h2.
  rewrite ih in h2.
  intro h3.
  destruct h3; contradiction.
  f_equal.
  rewrite ih.
  contradict h2.
  right; auto.
Qed.

Lemma inv_seg_fun_ge_bound_iff :
  forall {U n} (pfdu:type_eq_dec U)
         (f:seg_fun n U) y,
    inv_seg_fun pfdu y f >= n <->
    ~in_seg_fun y f.
  intros U n hdu f y; split; intros h1.
  intro h2.
  apply (inv_seg_fun_lt hdu) in h2.
  omega.
  rewrite <- (inv_seg_fun_eq_bound_iff hdu) in h1.
  rewrite h1; auto with arith.
Qed.

Lemma inv_seg_fun_compat :
  forall {U n} (pfdu:type_eq_dec U) (y:U) (f:seg_fun n U)
         (pf:inv_seg_fun pfdu y f < n),
    in_seg_fun_dec y f ->
    f (inv_seg_fun pfdu y f) pf = y.  
  intros U n.
  induction n as [|n ih].
  intros; omega.
  intros hdu y f. 
  specialize (ih hdu y (seg_fun_to_S f)).
  intro h1.
  simpl in h1.
  simpl.
  destruct (hdu (f 0 (lt_0_Sn n)) y) as [|h3]; subst.
  intros; f_equal.
  apply proof_irrelevance.
  pose proof (lt_S_n _ _ h1) as h5.
  intro h6.
  destruct h6 as [h6|h6]; subst.
  erewrite <- ih; auto.
  unfold seg_fun_to_S; f_equal; apply proof_irrelevance.
  contradict h3; auto.

  Grab Existential Variables.
  assumption.
Qed.


Lemma inv_seg_fun_compat_le :
  forall {U n} (pfdu:type_eq_dec U)
         (f:seg_fun n U) i (pfi:i < n),
    inv_seg_fun pfdu (f i pfi) f <= i.
  intros U n.
  induction n as [|n ih]; simpl; auto with arith.
  intros hdu f i hi.
  destruct i as [|i].
  erewrite f_equal_dep.
  rewrite eq_dec_same at 1.
  apply le_refl.
  assumption.
  reflexivity.
  specialize (ih hdu (seg_fun_to_S f) i (lt_S_n _ _ hi)).
  lr_if_goal; fold hlr; destruct hlr as [hl|hr]; auto with arith.
  apply le_n_S.
  unfold seg_fun_to_S in ih.
  erewrite f_equal_dep in ih.
  apply ih.
  reflexivity.
Qed.

  
  


Lemma inv_seg_fun_eq_iff :
  forall {U n} (pfdu:type_eq_dec U)
         (f:seg_fun n U) y k,
    inv_seg_fun pfdu y f = k <->
    (k = n /\ ~in_seg_fun y f) \/
    (exists (pfk:k < n),
       f k pfk = y /\
       (forall j (pfj:j < n),
          j < k ->
          f j pfj <> y)).
  intros U n hdu f y k.
  destruct (in_seg_fun_dec' hdu f y) as [h0 | h0].
  split; intro h2; subst. 
  destruct (lt_le_dec (inv_seg_fun hdu y f) n) as [h2 | h2].
  right. 
  exists h2; split.
  apply inv_seg_fun_compat; auto.
  apply in_seg_fun_impl_dec; auto.
  intros j hj.
  pose proof (inv_seg_fun_le hdu f j hj) as h4.
  intro h5.
  intro; subst.
  omega.
  apply le_ge in h2.
  rewrite inv_seg_fun_ge_bound_iff in h2.
  contradiction.
  destruct h2 as [h2|h2].
  destruct h2; contradiction. 
  destruct h2 as [h2 [? h4]]; subst.
  rewrite inv_seg_fun_eq_same_iff; auto. 
  split; intro h1; subst.
  left.
  split; auto.
  rewrite inv_seg_fun_eq_bound_iff; auto.
  destruct h1 as [h1 | h1].
  destruct h1; subst.
  rewrite inv_seg_fun_eq_bound_iff; auto. 
  destruct h1 as [h1 [? h2]]; subst.
  rewrite inv_seg_fun_eq_same_iff; auto.
Qed.

Lemma inv_seg_fun_eq0_iff :
  forall {U n} (pfdu:type_eq_dec U)
         (f:seg_fun n U) y (pf0:0 < n),
    inv_seg_fun pfdu y f = 0 <->
    y = f 0 pf0.
  intros; rewrite inv_seg_fun_eq_iff.
  split; intro h1.
  destruct h1 as [h1|h1].
  destruct h1; subst.
  omega.
  destruct h1 as [? [h1 h2]]; subst.
  f_equal; apply proof_irrelevance.
  right.
  subst.
  exists pf0.
  split; auto.
  intros; omega.
Qed.


Lemma inv_seg_fun_to_S_eq :
  forall {U n} (pfdu:type_eq_dec U) (f:seg_fun (S n) U) y,
    in_seg_fun_dec y (seg_fun_to_S f) ->
    y <> f 0 (lt_0_Sn _) ->
    inv_seg_fun pfdu y (seg_fun_to_S f) =
    pred (inv_seg_fun pfdu y f).
  intros U n hdu f y h2 hne.
  pose proof hne as hn.
  rewrite <- (inv_seg_fun_eq0_iff hdu) in hne.
  rewrite neq0_iff in hne.
  rewrite inv_seg_fun_eq_iff.
  right.
  ex_goal.
  pose proof (inv_seg_fun_lt hdu y (seg_fun_to_S f)
                             (in_seg_fun_dec_impl _ _ h2)) as h3.
  rewrite (S_pred _ _ hne).
  simpl.
  lr_if_goal; fold hlr; destruct hlr as [hl | hr]; subst.
  contradict hn; auto.
  simpl.
  apply inv_seg_fun_lt; auto.
  apply in_seg_fun_dec_impl; auto.
  exists hex.
  split.
  unfold seg_fun_to_S.
  gen0.
  rewrite <- (S_pred _ _ hne).
  intro h3.
  apply inv_seg_fun_compat.
  simpl.
  left; auto.
  intros j hj h3.
  unfold seg_fun_to_S. 
  intro; subst.
  pose proof (inv_seg_fun_weak_bound hdu f (S j) (lt_n_S _ _ hj)
                                     (f (S j) (lt_n_S _ _ hj))
                                     (eq_refl _)) as h5.
  omega.
Qed.

Lemma in_seg_fun_dec_lub_weak :
  forall {n} (f:seg_fun n nat),
    0 < n ->
    in_seg_fun_dec (seg_fun_lub_weak f) f.
  intro n.
  induction n as [|n ih].
  intros; omega.
  intros f h1.
  destruct (zerop n) as [|h2]; subst; simpl.
  right.
  destruct (zerop (f 0 (lt_0_Sn _))) as [h2 | h2].
  rewrite h2; auto.
  rewrite (S_pred _ _ h2); auto.
  specialize (ih (seg_fun_to_S f) h2). 
  destruct (lt_le_dec (seg_fun_lub_weak (seg_fun_to_S f)) (f 0 (lt_0_Sn n))) as [h3 | h3].
  apply lt_not_le in h3.
  apply nle_nge in h3.
  rewrite <- ge_nat_compat in h3.
  rewrite <- eq_false_iff in h3.
  rewrite h3; right; auto.
  apply le_ge in h3.
  rewrite <- ge_nat_compat in h3.
  rewrite h3.
  left; auto.
Qed.

Lemma in_seg_fun_lub_weak :
  forall {n} (f:seg_fun n nat),
    0 < n ->
    in_seg_fun (seg_fun_lub_weak f) f.
  intros; apply in_seg_fun_dec_impl; apply in_seg_fun_dec_lub_weak; auto.
Qed.


Lemma in_remove_seg_fun_dec :
  forall {U n} (pfdu:type_eq_dec U)
         m (f:seg_fun n U) y,
    in_seg_fun_dec y (remove_seg_fun m f) ->
    in_seg_fun_dec y f.
  intros U n hdu m f y h1; apply in_seg_fun_impl_dec; apply in_seg_fun_dec_impl in h1; auto.
  rewrite in_seg_fun_iff in h1.
  destruct h1 as [i [h1 ?]]; subst.
  rewrite in_seg_fun_iff. 
  destruct (lt_le_dec i m) as [h2 | h2].
  exists i, (lt_pred' _ _ h1).
  rewrite remove_seg_fun_lt; auto.
  exists (S i), (lt_pred_S _ _ h1).
  rewrite remove_seg_fun_ge; auto.
Qed.

Lemma in_remove_seg_fun : 
  forall {U n}
         m (f:seg_fun n U) y,
    in_seg_fun y (remove_seg_fun m f) ->
    in_seg_fun y f.
  intros U n m f y h1.
  rewrite in_seg_fun_iff in h1.
  destruct h1 as [i [h1 ?]]; subst.
  rewrite in_seg_fun_iff. 
  destruct (lt_le_dec i m) as [h2 | h2].
  exists i, (lt_pred' _ _ h1).
  rewrite remove_seg_fun_lt; auto.
  exists (S i), (lt_pred_S _ _ h1).
  rewrite remove_seg_fun_ge; auto.
Qed.


(*Same as [remove_seg_fun], only you specify an element in the image.*)
(*Only of course removes the first instance of [y]*)
Fixpoint remove_im_seg_fun {U n} (pfdu:type_eq_dec U) (y:U) :
  forall (f:seg_fun n U), in_seg_fun_dec y f -> seg_fun (pred n) U :=
  match n with
    | 0 => fun f pfin => False_rect _ pfin
    | S n' =>
      fun f pfin =>
        let f' := seg_fun_to_S f in
        match pfin with
          | inleft pfin' =>
            fun i =>
              match i with
                | 0 => 
                  fun pf0 =>
                    if (pfdu (f 0 (lt_0_Sn _)) y) then
                      f' 0 pf0 
                    else
                      f 0 (lt_0_Sn _)
                | S i' =>
                  fun pfi => 
                    let pfi' :=  lt_S_pred _ _ pfi in
                    let pfi'' := lt_n_S _ _ pfi in
                    if (pfdu (f 0 (lt_0_Sn _)) y) then
                      f (S (S i')) pfi''
                    else
                      remove_im_seg_fun pfdu y f' pfin' i' pfi'
              end
          | inright pfe =>
            f'
        end
  end.


Lemma remove_im_seg_fun_lt :
  forall {U n} (pfdu:type_eq_dec U) (y:U) (f:seg_fun n U)
         (pfin:in_seg_fun_dec y f)
         i
         (pfi:i < pred n)
         (pfi':i < n),
      i < inv_seg_fun pfdu y f ->
      remove_im_seg_fun pfdu y f pfin i pfi = f i pfi'.
  intros U n.  
  induction n as [|n ih]; simpl.
  intros; omega.
  intros hdu y f hin i hi hi'. 
  lr_if_goal; fold hlr; destruct hlr as [hl | hr]; subst.
  intros; omega.  
  destruct hin as [h1 | h1].
  destruct i as [|i].
  intro.
  f_equal; apply proof_irrelevance.
  intro h2.
  apply lt_S_n in h2.
  specialize (ih hdu y (seg_fun_to_S f) h1 _ (lt_S_pred _ _ hi) (lt_S_n _ _ hi') h2).
  rewrite ih.
  unfold seg_fun_to_S.
  f_equal; apply proof_irrelevance.
  contradiction.
Qed.


Lemma remove_im_seg_fun_ge :
  forall {U n} (pfdu:type_eq_dec U) (y:U) (f:seg_fun n U)
         (pfin:in_seg_fun_dec y f)
         i
         (pfi:i < pred n),
    i >= inv_seg_fun pfdu y f ->
    let pfi' := lt_pred_S _ _ pfi in
    remove_im_seg_fun pfdu y f pfin i pfi = f (S i) pfi'.
  intros U n.      
  induction n as [|n ih]; simpl.
  intros; omega. 
  intros hdu y f hin i hi.
  intro h2.
  destruct hin as [h1 | h1].
  destruct i as [|i].
  lr_if h2; fold hlr in h2; destruct hlr as [hl | hr]; subst.
  rewrite eq_dec_same at 1.
  unfold seg_fun_to_S.
  f_equal; apply proof_irrelevance.
  assumption.
  omega.   
  lr_if h2; fold hlr in h2; destruct hlr as [hl | hr]; subst.
  rewrite eq_dec_same at 1.
  f_equal.
  apply proof_irrelevance.
  assumption.
  apply le_S_n in h2.
  lr_if_goal; fold hlr; destruct hlr as [hl' | hr']; subst.
  f_equal.
  apply proof_irrelevance.
  rewrite ih; auto.
  unfold seg_fun_to_S.
  f_equal; apply proof_irrelevance.
  unfold seg_fun_to_S.
  f_equal; apply proof_irrelevance.
Qed.


Lemma remove_im_seg_fun_eq_same_iff :
  forall {U n} (pfdu:type_eq_dec U)
         (f:seg_fun (S n) U) y i (pfi:i < n)
         (pfin:in_seg_fun_dec y f),
    remove_im_seg_fun pfdu y f pfin i pfi = y <->
    (i >= inv_seg_fun pfdu y f /\ f (S i) (lt_n_S _ _ pfi) = y).
  intros U n.
  induction n as [|n ih]; simpl.
  intros; omega.
  intros hdu f y i hi.
  destruct i as [|i]; simpl.
  intro hin. 
  destruct hin as [[h2|h2]|n2].
  lr_if_goal; fold hlr; destruct hlr as [hl|hr].
  unfold seg_fun_to_S; intuition.
  split; intro h3; subst.
  contradict hr; auto.
  omega.
  subst.
  lr_if_goal; fold hlr; destruct hlr as [hl|hr].
  intuition.
  split; intro;[contradiction|omega].
  subst.
  rewrite eq_dec_same at 1.
  unfold seg_fun_to_S.
  intuition.
  assumption.
  specialize (ih hdu (seg_fun_to_S f) y i (lt_S_n _ _ hi)).
  intro hin.
  destruct hin as [h2|h2].
  specialize (ih h2).
  lr_if_goal; fold hlr; destruct hlr as [hl|hr].
  intuition.
  simpl in ih.
  destruct h2 as [h2|h2], i as [|i]. 
  lr_if_goal; fold hlr; fold hlr in ih; destruct hlr as [hl'|hr']; subst.
  split; intro h3.
  unfold seg_fun_to_S in hr, h3.
  split; auto with arith.
  unfold seg_fun_to_S. rewrite <- h3.
  f_equal; apply proof_irrelevance.
  destruct h3 as [h3 h4].
  unfold seg_fun_to_S in h4; unfold seg_fun_to_S.
  rewrite <- h4.
  f_equal; apply proof_irrelevance.
  split; intro; subst.
  pose proof (iff1 ih (eq_refl _)); omega.
  destruct H as [h3 h4]; subst.
  omega. 
  lr_if_goal; fold hlr; fold hlr in ih; destruct hlr as [hl'|hr']; subst.
  erewrite f_equal_dep.
  rewrite ih at 1.
  split; intro h3; destruct h3 as [h3 h4]; split; auto with arith.
  rewrite <- h4.
  unfold seg_fun_to_S; f_equal; apply proof_irrelevance.
  unfold seg_fun_to_S in h4; unfold seg_fun_to_S.
  rewrite <- h4; f_equal; apply proof_irrelevance.
  reflexivity.
  erewrite f_equal_dep.
  rewrite ih at 1.
  split; intro h3; destruct h3 as [h3 h4]; subst.
  split.
  apply le_n_S; auto.
  unfold seg_fun_to_S; f_equal; apply proof_irrelevance.
  split.
  eapply lt_S_n in h3; auto.
  unfold seg_fun_to_S; f_equal; apply proof_irrelevance.
  reflexivity.
  subst.
  rewrite eq_dec_same in ih at 1.
  rewrite eq_dec_same at 1.
  simpl in ih.
  erewrite f_equal_dep.
  rewrite ih at 1.
  split; intro h1; destruct h1 as [h1 h2]; unfold seg_fun_to_S in h2; unfold seg_fun_to_S.
  split; auto with arith.
  rewrite <- h2; f_equal; apply proof_irrelevance.
  split; auto.
  rewrite <- h2; f_equal; apply proof_irrelevance.
  reflexivity.
  assumption.
  assumption.
  subst.
  rewrite eq_dec_same in ih at 1.
  rewrite eq_dec_same at 1; simpl in ih.
  erewrite f_equal_dep.
  rewrite ih at 1.
  split; intro h1; destruct h1 as [h1 h2]; unfold seg_fun_to_S in h2; unfold seg_fun_to_S.
  split; auto with arith.
  rewrite <- h2; f_equal; apply proof_irrelevance.
  split; auto with arith.
  rewrite <- h2; f_equal; apply proof_irrelevance.
  reflexivity.
  assumption.
  assumption.
  subst.
  rewrite eq_dec_same at 1.
  split; intro h2; destruct h2 as [h2].
  split; auto with arith.
  unfold seg_fun_to_S; auto.
  assumption.
Qed.

Lemma remove_im_seg_fun_in_nin :
  forall {U n} (pfdu:type_eq_dec U)
         (f:seg_fun n U) y b
         (pfin:in_seg_fun_dec y f),
    in_seg_fun b f ->
    ~in_seg_fun b (remove_im_seg_fun pfdu y f pfin) ->
    b = y.
  intros U n hdu f y b hin hin' h2.
  rewrite in_seg_fun_iff in hin'.
  destruct hin' as[i [hi ?]]; subst.
  pose proof (in_seg_fun_dec_impl _ _ hin) as hj.
  rewrite in_seg_fun_iff in hj.
  destruct hj as [j [hj ?]]; subst.
  destruct i as [|i].
  destruct (zerop (inv_seg_fun hdu (f j hj) f)) as [h5|h5].
  rewrite inv_seg_fun_eq0_iff in h5.
  rewrite <- h5 at 1.
  reflexivity.
  destruct (zerop j) as [|h3]; subst.
  f_equal; apply proof_irrelevance.
  destruct (hdu (f 0 hi) (f j hj)) as [h4|h4]; auto.
  contradict h2.
  rewrite in_seg_fun_iff.
  exists 0.
  ex_goal.
  omega. 
  exists hex.
  erewrite remove_im_seg_fun_lt; auto.
  destruct (zerop (inv_seg_fun hdu (f (S i) hi) f)) as [h5|h5].  
  rewrite inv_seg_fun_eq0_iff in h5.
  destruct j as [|j].
  rewrite h5. 
  f_equal; apply proof_irrelevance.
  rewrite h5 in h2.
  destruct (hdu (f (S i) hi) (f (S j) hj)) as [|h6]; auto.
  rewrite h5 in h6.
  contradict h2.  
  rewrite in_seg_fun_iff. 
  exists 0.
  ex_goal.
  destruct n as [|n].
  omega.
  simpl; omega.
  exists hex. 
  erewrite remove_im_seg_fun_lt.
  reflexivity.  
  destruct (zerop (inv_seg_fun hdu (f (S j) hj) f)) as [h7|]; auto. 
  rewrite inv_seg_fun_eq0_iff in h7.
  rewrite <- h7 in h6 at 1.
  contradict h6; auto.
  destruct (hdu (f (S i) hi) (f j hj)) as [|h6]; auto.
  contradict h2.
  rewrite in_seg_fun_iff.
  destruct (lt_le_dec (S i) (inv_seg_fun hdu (f j hj) f)) as [h7|h7].
  exists (S i).
  ex_goal.
  pose proof (inv_seg_fun_lt hdu (f j hj) f) as h8.
  hfirst h8.
  apply in_seg_fun_compat.
  specialize (h8 hf).
  omega.
  exists hex. 
  erewrite remove_im_seg_fun_lt; auto.
  apply le_dec' in h7.
  destruct h7 as [h7|h7].
  exists i.
  ex_goal.
  omega.
  exists hex.
  erewrite remove_im_seg_fun_ge; auto.
  f_equal; apply proof_irrelevance.
  contradict h6.
  gen0.
  rewrite <- h7.
  intro.
  rewrite inv_seg_fun_compat; auto.

  Grab Existential Variables.
  omega.
Qed.



Lemma inj_seg_fun_eq_iff_inv :
  forall {U n} (pfdu:type_eq_dec U)
         (f:seg_fun n U) y ,
    inj_seg f ->
    forall i (pf:i < n),
      f i pf = y <->
      (exists _:in_seg_fun_dec y f,
       i = inv_seg_fun pfdu y f).
  intros U n hdu f y h1 i h2; split; intro h4; subst. 
  exists (in_seg_fun_dec_compat _ _ _ ).
  pose proof (inv_seg_fun_le hdu f i h2) as h3.
  apply le_lt_eq_dec in h3.
  destruct h3 as [h3 | h3]; auto.
  apply lt_neq in h3.
  rewrite (inv_seg_fun_eq_same_iff _ _) in h3.
  contradict h3.
  intros j hj h3 h4.
  apply h1 in h4; subst; contradict (lt_irrefl _ hj).
  destruct h4 as [h4]; subst.
  apply in_seg_fun_dec_impl in h4.
  rewrite in_seg_fun_iff in h4.
  destruct h4 as [i [h4 ?]]; subst.
  rewrite inv_seg_fun_compat; auto.
  apply in_seg_fun_dec_compat.
Qed.
   
Lemma inj_seg_fun_count_compat :
  forall {n} (f:seg_fun n nat) r,
    inj_seg_fun_endo f r ->
    forall y,
      in_seg_fun_dec y f ->
      count_seg_fun_im_nat y f = 1.
  intro n. 
  induction n as [|n ih]; simpl.
  intros; contradiction.
  intros f r h1 y h2.
  destruct (nat_eq_dec y (f 0 (lt_0_Sn n))) as [|h3]; subst.
  rewrite beq_nat_same.
  f_equal. 
  apply le0.
  apply not_lt_le.
  apply ngt_nlt.
  rewrite count_seg_fun_im_in_nat_iff.
  intro h3.
  rewrite in_seg_fun_iff in h3.
  destruct h3 as [i [h3 h4]].
  unfold seg_fun_to_S in h4.
  inversion h1 as [? ? h5 h6]; subst.
  apply h6 in h4.
  discriminate.
  pose proof h3 as h3'.
  erewrite <- neq_nat_compat in h3.
  rewrite h3.
  eapply ih.
  apply inj_seg_fun_endo_to_S in h1.
  apply h1.
  destruct h2; subst; auto.
  contradict h3'; auto.
Qed.



Lemma lt_inv_seg_fun_lub_weak :
  forall {n} (f:seg_fun n nat) (pf0:0 < n),
    inv_seg_fun_lub_weak f pf0 < n.
  unfold inv_seg_fun_lub_weak; intros; apply inv_seg_fun_dec_lt;
  apply in_seg_fun_dec_lub_weak; auto.
Qed.


Lemma inv_seg_fun_lub_weak_compat :
  forall {n}
         (f:seg_fun n nat) (pf0:0 < n)
         (pf:inv_seg_fun_lub_weak f pf0 < n),
    f (inv_seg_fun_lub_weak f pf0) pf =
    seg_fun_lub_weak f.
  unfold inv_seg_fun_lub_weak; intros n f h0 h1.
  pose proof (inv_seg_fun_compat nat_eq_dec (seg_fun_lub_weak f) f h1) as h2.
  rewrite <- h2.
  apply f_equal_dep; f_equal.
  apply in_seg_fun_dec_lub_weak; auto.
Qed.


Lemma le_seg_fun_lub_weak_from_S :
  forall {n} (f:seg_fun (S n) nat),
    seg_fun_lub_weak (seg_fun_from_S f) <=
    seg_fun_lub_weak f.
  intros n f.
  destruct (zerop n) as [|hn]; subst.  
  simpl; auto with arith.
  pose proof (seg_fun_lub_compat f (lt_0_Sn _)) as h6.
  destruct h6 as [h6 h7].
  specialize (h6 (inv_seg_fun_lub_weak (seg_fun_from_S f) hn)).
  hfirst h6.
  apply lt_S.
  apply lt_inv_seg_fun_lub_weak.
  specialize (h6 hf).
  pose proof (inv_seg_fun_lub_weak_compat (seg_fun_from_S f) hn
                                          (lt_inv_seg_fun_lub_weak _ _)) as h8.
  rewrite <- h8.
  rewrite seg_fun_from_S_compat.
  unfold seg_fun_lub in h6.
  apply lt_dec in h6.
  destruct h6 as [h6 | h6].
  erewrite f_equal_dep.
  pose proof (lt_eq_trans _ _ _ h6 (eq_refl (seg_fun_lub_weak f))) as h9.
  eapply lt_le_weak.
  apply h9.
  reflexivity. 
  erewrite f_equal_dep. 
  rewrite h6 at 1.
  auto.
  reflexivity.
Qed.


Lemma le_seg_fun_lub_weak_remove :
  forall {n} (f:seg_fun n nat) i,
    seg_fun_lub_weak (remove_seg_fun i f) <=
    seg_fun_lub_weak f.
  intros n f i.
  destruct (zerop n) as [|h1]; subst.
  simpl; auto.
  pose proof (seg_fun_lub_compat f h1) as h2.
  destruct h2 as [h2 h3].
  destruct (zerop (pred n)) as [h4|h4].
  apply pred0_dec in h4.
  destruct h4; subst; auto with arith.
  specialize (h2 (inv_seg_fun nat_eq_dec (seg_fun_lub_weak (remove_seg_fun i f)) f)).
  hfirst h2.
  apply inv_seg_fun_lt.
  pose proof (in_seg_fun_lub_weak (remove_seg_fun i f) h4) as h5.
  apply in_remove_seg_fun in h5; auto. 
  specialize (h2 hf).
  rewrite inv_seg_fun_compat in h2.
  unfold seg_fun_lub in h2; auto with arith.
  pose proof (in_seg_fun_lub_weak (remove_seg_fun i f) h4) as h5.
  apply in_remove_seg_fun in h5; auto.
  apply in_seg_fun_impl_dec; auto.
  apply nat_eq_dec.
Qed.
                 

Lemma seg_fun_endo_remove_im :
  forall {n i m} (f:seg_fun n nat),
    i < n ->
    seg_fun_endo f m -> 
    forall (pfi:in_seg_fun_dec i f),
      seg_fun_endo (remove_im_seg_fun nat_eq_dec i f pfi) m.
  intros n i m f h1 h2 h3.
  inversion h2; subst.
  constructor.
  intros j h4.
  destruct (lt_le_dec j (inv_seg_fun nat_eq_dec i f)) as [h5 | h5].
  erewrite remove_im_seg_fun_lt; auto.
  erewrite remove_im_seg_fun_ge; auto.

  Grab Existential Variables.
  omega.
Qed.



Lemma seg_fun_endo_remove_im' :
  forall {n} (f:seg_fun n nat)
         a b c
         (pfi:in_seg_fun_dec a f),
    b <= c ->
    a < c ->
    seg_fun_endo (remove_im_seg_fun nat_eq_dec a f pfi) b ->
    seg_fun_endo f c.
  intros n f a b c hin h2 h3 h4.
  inversion h4 as [? ? h5 h6]; subst.
  destruct (zerop n) as [|hz]; subst.
  constructor.
  intros i hi.
  contradict (lt_n_0 _ hi). 
  constructor.
  intros i hi.
  destruct (lt_dec _ _ hi) as [h6 | h6].
  destruct (lt_le_dec i (inv_seg_fun nat_eq_dec a f)) as [h7 | h7].
  specialize (h5 i h6). 
  erewrite remove_im_seg_fun_lt in h5; auto.
  eapply lt_le_trans.
  apply h5.  
  assumption.   
  apply le_lt_eq_dec in h7.
  destruct h7 as [h7|h7].
  destruct i as [|i].
  omega.
  apply lt_le in h7.
  specialize (h5 i (lt_trans _ _ _ (lt_n_Sn _) h6)).
  rewrite remove_im_seg_fun_ge in h5; auto.
  eapply lt_le_trans.
  erewrite f_equal_dep.
  apply h5.
  reflexivity.
  assumption.
  subst. 
  rewrite inv_seg_fun_compat; auto.
  destruct (lt_le_dec i (inv_seg_fun nat_eq_dec a f)) as [h7 | h7].
  subst.
  pose proof (inv_seg_fun_lt nat_eq_dec a f) as h8.
  hfirst h8.
  apply in_seg_fun_dec_impl; auto.
  specialize (h8 hf).
  omega. 
  subst.
  destruct (zerop (pred n)) as [h6|h6].
  apply pred0_dec in h6; destruct h6; subst.
  contradict (lt_irrefl _ hi). 
  simpl.
  simpl in h7.
  apply le0 in h7.
  lr_if h7; fold hlr in h7; destruct hlr as [hl | hr].
  subst.
  erewrite f_equal_dep.
  apply h3.
  reflexivity.
  discriminate.
  specialize (h5 (pred (pred n))).
  hfirst h5.
  eapply lt_pred_n.
  apply h6.
  apply le_dec' in h7.
  destruct h7 as [h7 | h7].
  specialize (h5 hf).
  rewrite remove_im_seg_fun_ge in h5; auto.
  generalize hi.
  rewrite (S_pred _ _ h6).
  intro h8.
  erewrite f_equal_dep.
  eapply lt_le_trans.
  apply h5.
  assumption.
  reflexivity.
  rewrite inv_seg_fun_eq_iff in h7.
  destruct h7 as [h7 | h7].
  destruct h7 as [h7].
  contradict (lt_irrefl _  (eq_lt_trans _ _ _ (eq_sym h7) hi)).
  destruct h7 as [h7 [h8 ?]]; subst.
  erewrite f_equal_dep.
  apply h3.
  reflexivity.
Qed.

              
Lemma inj_seg_fun_endo_remove_im :
  forall {n i m} (f:seg_fun n nat),
    i < n ->
    inj_seg_fun_endo f m -> 
    forall (pfi:in_seg_fun_dec i f),
      inj_seg_fun_endo (remove_im_seg_fun nat_eq_dec i f pfi) m.
  intros n i m f h1 h2 h3. 
  inversion h2; subst.
  constructor.
  apply seg_fun_endo_remove_im; auto.
  intros r s hr hs h4.
  destruct
    (lt_le_dec r (inv_seg_fun nat_eq_dec i f))
    as [h5 | h5],
       (lt_le_dec s (inv_seg_fun nat_eq_dec i f)) as [h6 | h6].
  pose proof @remove_im_seg_fun_lt.
  rewrite (remove_im_seg_fun_lt nat_eq_dec i f h3 r hr (lt_pred' _ _ hr)),
  (remove_im_seg_fun_lt nat_eq_dec i f h3 s hs (lt_pred' _ _ hs)) in h4.
  apply H0 in h4; auto.
  rewrite (remove_im_seg_fun_lt nat_eq_dec i f h3 s hs (lt_pred' _ _ hs)) in h4; auto.
  auto. 
  rewrite (remove_im_seg_fun_lt nat_eq_dec i f h3 r hr (lt_pred' _ _ hr)),
  (remove_im_seg_fun_ge nat_eq_dec i f h3 s hs) in h4; auto.
  apply H0 in h4; subst; omega.
  rewrite (remove_im_seg_fun_ge nat_eq_dec i f h3 r),
  (remove_im_seg_fun_lt nat_eq_dec i f h3 s hs (lt_pred' _ _ hs)) in h4; auto.
  apply H0 in h4; subst; omega.
  rewrite (remove_im_seg_fun_ge nat_eq_dec i f h3 s hs),
  (remove_im_seg_fun_ge nat_eq_dec i f h3 r hr) in h4; auto.
  apply H0 in h4.
  apply S_inj in h4; auto.
Qed.


Lemma inj_seg_fun_endo_remove_im' :
  forall {n} (f:seg_fun n nat)
         a b c
         (pfin:in_seg_fun_dec a f),
    ~in_seg_fun a (remove_im_seg_fun nat_eq_dec a f pfin) ->
    b <= c ->
    a < c ->
    inj_seg_fun_endo (remove_im_seg_fun nat_eq_dec a f pfin) b ->
    inj_seg_fun_endo f c.
  intros n f a b c hin hnin hb ha h1. 
  inversion h1 as [? ? h2 h3]; subst.
  constructor.
  eapply seg_fun_endo_remove_im'.
  apply hb.
  apply ha.
  apply h2.
   
  destruct (zerop (pred n)) as [h4|h4]; subst.
  apply pred0_dec in h4; destruct h4; subst.
  intros; omega.
  intros; omega.  
  pose proof (in_seg_fun_dec_impl _ _ hin) as hin'.
  intros i j hi hj h5.
  destruct (lt_dec _ _ hi) as [h6|h6], (lt_dec _ _ hj) as [h7|h7].
    destruct
    (lt_le_dec i (inv_seg_fun nat_eq_dec a f))
    as [h8|h8],
       (lt_le_dec j (inv_seg_fun nat_eq_dec a f))
      as [h9|h9].
  specialize (h3 _ _ h6 h7).
  apply h3.
  erewrite remove_im_seg_fun_lt at 1.
  erewrite remove_im_seg_fun_lt at 1.
  apply h5.
  assumption. 
  assumption.
  destruct (zerop j) as [|hz]; subst.
  apply le0 in h9.
  omega.
  apply le_dec' in h9.
  destruct h9 as [h9|h9].
  specialize (h3 i (pred j) h6).
  hfirst h3.
  apply lt_S_n.
  rewrite <- (S_pred _ _ hz).
  rewrite <- (S_pred _ _ (Olt _ _ hi)); auto.
  specialize (h3 hf).
  hfirst h3.
  erewrite remove_im_seg_fun_lt; auto.
  erewrite remove_im_seg_fun_ge; auto.
  symmetry.
  gen0.
  rewrite <- (S_pred _ _ hz).
  intro h10.
  erewrite h5 at 1.
  f_equal; apply proof_irrelevance.
  apply h3 in hf0; subst.
  pose proof (lt_le_trans _ _ _ h8 h9) as h10.
  apply lt_irrefl in h10; contradiction.
  subst.
  rewrite inv_seg_fun_compat in h5; auto; subst.
  contradict hnin.
  rewrite in_seg_fun_iff.
  exists i, h6.
  erewrite remove_im_seg_fun_lt; auto.
  destruct (zerop i) as [|hz]; subst.
  apply le0 in h8.
  omega.
  apply le_dec' in h8.
  destruct h8 as [h8|h8].
  specialize (h3 (pred i) j).
  hfirst h3.
  eapply lt_trans.
  apply lt_pred.
  rewrite <- (S_pred _ _ hz).
  apply h6.
  eapply lt_pred_n.
  apply h6.
  specialize (h3 hf h7).
  hfirst h3.
  erewrite remove_im_seg_fun_ge; auto.
  erewrite remove_im_seg_fun_lt; auto.
  gen0.
  rewrite <- (S_pred _ _ hz).
  intro.
  erewrite <- h5 at 1.
  f_equal; apply proof_irrelevance.
  specialize (h3 hf0); subst. 
  pose proof (lt_le_trans _ _ _ h9 h8) as h10.
  apply lt_irrefl in h10; contradiction.
  subst. 
  rewrite inv_seg_fun_compat in h5; subst; auto.
  pose proof (inv_seg_fun_weak_bound nat_eq_dec f j hj _ (eq_refl _)) as h10.
  pose proof (lt_le_trans _ _ _ h9 h10) as h11.
  contradict (lt_irrefl _ h11).
  destruct (zerop i) as [|h10], (zerop j) as [|h11]; subst; auto.
  apply le0 in h8.
  rewrite inv_seg_fun_eq0_iff in h8.
  rewrite h5 in h8 at 1.
  subst.
  rewrite <- h5 in hnin at 1.
  contradict hnin.
  rewrite in_seg_fun_iff.
  apply le_dec' in h9.
  destruct h9 as [h9|h9].
  exists (pred j).
  ex_goal.
  omega.
  exists hex.
  rewrite remove_im_seg_fun_ge; auto.
  gen0.
  rewrite <- (S_pred _ _ h11).
  intro.
  rewrite h5. 
  f_equal; apply proof_irrelevance. 
  rewrite inv_seg_fun_eq_same_iff in h9.
  specialize (h9 0 h11 hi).
  rewrite h5 in h9.
  contradict h9; auto.
  apply le0 in h9.
  rewrite inv_seg_fun_eq0_iff in h9.
  rewrite <- h5 in h9 at 1.
  subst.
  apply le_dec' in h8.
  destruct h8 as [h8|h8].
  contradict hnin.  
  rewrite in_seg_fun_iff.
  exists (pred i).
  ex_goal.
  omega.
  exists hex.
  rewrite remove_im_seg_fun_ge; auto.
  gen0.
  rewrite <- (S_pred _ _ h10).
  intros; f_equal.
  apply proof_irrelevance.
  rewrite inv_seg_fun_eq_same_iff in h8.
  specialize (h8 0 h10 hj).
  rewrite <- h5 in h8.
  contradict h8; auto.
  apply le_dec' in h8.
  destruct h8 as [h8|h8].
  apply le_dec' in h9.
  destruct h9 as [h9|h9].
  specialize (h3 (pred i) (pred j)).
  hfirst h3.
  eapply lt_trans.
  eapply lt_pred_n.
  apply h10.
  assumption.
  specialize (h3 hf).
  hfirst h3.
  eapply lt_trans.
  eapply lt_pred_n.
  apply h11.
  assumption.
  specialize (h3 hf0).
  hfirst h3.
  rewrite remove_im_seg_fun_ge; auto.
  rewrite remove_im_seg_fun_ge; auto.
  gen0.
  rewrite <- (S_pred _ _ h10).
  intros; symmetry.
  gen0.
  rewrite <- (S_pred _ _ h11).
  intros.
  erewrite f_equal_dep.
  rewrite <- h5 at 1.
  f_equal; apply proof_irrelevance.
  reflexivity.
  specialize (h3 hf1).
  rewrite (S_pred _ _ h10).
  rewrite h3.
  symmetry.
  eapply S_pred.
  apply h11.
  subst.
  rewrite inv_seg_fun_compat in h5; subst; auto.
  contradict hnin.
  rewrite in_seg_fun_iff.
  exists (pred i).
  ex_goal.
  omega.
  exists hex.
  rewrite remove_im_seg_fun_ge; auto.
  gen0.
  rewrite <- (S_pred _ _ h10).
  intros; f_equal; apply proof_irrelevance.
  subst.
  rewrite inv_seg_fun_compat in h5; subst. 
  apply le_dec' in h9.
  destruct h9 as [h9|]; auto.
  contradict hnin.
  rewrite in_seg_fun_iff.
  exists (pred j).
  ex_goal.
  omega. 
  exists hex.
  rewrite remove_im_seg_fun_ge; auto.
  gen0.
  rewrite <- (S_pred _ _ h11).
  intro; f_equal; apply proof_irrelevance.
  assumption.
  subst. 
  pose proof (inv_seg_fun_lt nat_eq_dec a f hin') as h7. 
  apply lt_le_pred in h7.
  apply le_dec' in h7.
  destruct (lt_le_dec i (inv_seg_fun nat_eq_dec a f)) as [h8|h8].
  destruct h7 as [h7|h7].
  specialize (h3 i (pred (pred n)) h6).
  hfirst h3.
  eapply lt_pred_n.
  apply h6.
  specialize (h3 hf).
  hfirst h3.
  erewrite remove_im_seg_fun_lt; auto.
  erewrite remove_im_seg_fun_ge; auto.
  symmetry.
  gen0.
  rewrite <- (S_pred _ _ h4).
  intro.
  rewrite h5 at 1.
  f_equal; apply proof_irrelevance.
  specialize (h3 hf0).
  subst.
  pose proof (lt_le_trans _ _ _ h8 h7) as h9.
  apply lt_irrefl in h9; contradiction.
  rewrite inv_seg_fun_eq_iff in h7.
  destruct h7 as [h7 | h7].
  destruct h7; contradiction.
  destruct h7 as [? [h7 h9]]; subst.
  specialize (h9 i hi h6).
  rewrite h5 in h9. 
  contradict h9; f_equal; apply proof_irrelevance.
  destruct (zerop i) as [|hz]; subst.
  apply le0 in h8.
  rewrite inv_seg_fun_eq0_iff in h8.
  rewrite h5 in h8 at 1.
  subst.
  contradict hnin.
  rewrite in_seg_fun_iff.
  destruct h7 as [h7|h7].
  exists (pred (pred n)).
  ex_goal.
  eapply lt_pred_n.
  apply h6.
  exists hex.
  rewrite remove_im_seg_fun_ge; auto.
  gen0.
  rewrite <- (S_pred _ _ h6).
  intro; f_equal; apply proof_irrelevance.
  rewrite inv_seg_fun_eq_iff in h7.
  destruct h7 as [h7|h7].
  destruct h7; contradiction.
  destruct h7 as [h7 [h8 h9]].
  specialize (h9 0 hi h6).
  rewrite h5 in h9.
  contradict h9; auto.
  apply le_dec' in h8.
  destruct h7 as [h7|h7].
  destruct h8 as [h8|h8]. 
  specialize (h3 (pred i) (pred (pred n))). 
  hfirst h3.
  eapply lt_trans.
  apply lt_n_Sn.
  rewrite <- (S_pred _ _ hz).
  assumption.
  specialize (h3 hf).
  hfirst h3.
  eapply lt_pred_n.
  apply h6.
  specialize (h3 hf0).
  hfirst h3.
  rewrite remove_im_seg_fun_ge; auto.
  rewrite remove_im_seg_fun_ge; auto.
  gen0.
  rewrite <- (S_pred _ _ hz).
  intro.
  symmetry.
  gen0.
  rewrite <- (S_pred _ _ h6).
  intro.
  erewrite f_equal_dep.
  rewrite <- h5 at 1.
  f_equal; apply proof_irrelevance.
  reflexivity.
  specialize (h3 hf1).
  rewrite (S_pred _ _ hz).
  rewrite h3.
  rewrite <- (S_pred _ _ h6); auto.
  subst.
  rewrite inv_seg_fun_compat in h5; subst.
  rewrite inv_seg_fun_eq_same_iff.
  intros j hj'  h9 h10.
  destruct (lt_le_dec j (inv_seg_fun nat_eq_dec (f (pred n) hj) f)) as [h11|h11].
  contradict hnin.
  rewrite in_seg_fun_iff.
  exists j, hj'.
  erewrite remove_im_seg_fun_lt; auto.
  rewrite <- h10.
  reflexivity.
  specialize (h3 (pred (pred n)) (pred j)).
  hfirst h3.  
  eapply lt_pred_n.
  apply hj'.
  specialize (h3 hf). 
  hfirst h3.
  pose proof (inv_seg_fun_eq0_iff nat_eq_dec f (f (pred n) hj) (Olt _ _ h9)) as h12.
  nhl h12. 
  rewrite neq0_iff.
  assumption. 
  rewrite h12 in hl.
  destruct (zerop j) as [|h13]; subst; auto.
  rewrite (S_pred _ _ h13).
  eapply mono_pred_lt.
  rewrite <- (S_pred _ _ h13); auto.
  auto with arith.
  specialize (h3 hf0).
  destruct (zerop j) as [|h12]; subst.
  contradict hnin.
  rewrite in_seg_fun_iff.
  exists _, hf.
  rewrite remove_im_seg_fun_ge; auto.
  gen0.
  rewrite <- (S_pred _ _ hj').
  intro; f_equal; apply proof_irrelevance.
  apply le_dec' in h11. 
  destruct h11 as [h11|h11].
  hfirst h3.
  erewrite remove_im_seg_fun_ge; auto.
  rewrite remove_im_seg_fun_ge; auto. 
  gen0. 
  rewrite <- (S_pred _ _ h4).
  intro; symmetry; gen0; rewrite <- (S_pred _ _ h12).
  intro.
  erewrite f_equal_dep.
  rewrite h10 at 1.
  f_equal; apply proof_irrelevance.
  reflexivity. 
  specialize (h3 hf1). 
  pose proof (f_equal S h3) as h13.
  rewrite <- (S_pred _ _ h4), <- (S_pred _ _ h12) in h13; subst.
  apply lt_irrefl in hj'; auto. 
  subst.
  contradict hnin.
  rewrite in_seg_fun_iff.
  exists _, (lt_pred_n _ _ hj').
  rewrite remove_im_seg_fun_ge; auto.
  gen0.
  rewrite <- (S_pred _ _ h4).
  intro; f_equal; apply proof_irrelevance.
  assumption.
  pose proof (f_equal_dep f _ _ (inv_seg_fun_lt nat_eq_dec a f hin') hj h7) as h9.
  rewrite inv_seg_fun_compat in h9; subst.
  rewrite h7 in h8.
  destruct h8; omega.
  assumption. 
  subst.
  pose proof (inv_seg_fun_lt nat_eq_dec a f hin') as h9.
  apply lt_le_pred in h9.
  apply le_dec' in h9.
  destruct h9 as [h9|h9].
  destruct (lt_le_dec j (inv_seg_fun nat_eq_dec a f)) as [h8 | h8].
  specialize (h3 j (pred (pred n)) h7).
  hfirst h3.
  eapply lt_pred_n.
  apply h7.
  specialize (h3 hf).
  hfirst h3.
  erewrite remove_im_seg_fun_lt; auto.
  erewrite remove_im_seg_fun_ge; auto.
  symmetry.
  gen0.
  rewrite <- (S_pred _ _ h7).
  intro.
  rewrite <- h5 at 1.
  f_equal; apply proof_irrelevance.
  specialize (h3 hf0).
  subst. 
  pose proof (lt_le_trans _ _ _ h8 h9) as h10.
  contradict (lt_irrefl _ h10).
  destruct (zerop j) as [|h10]; subst.
  apply le0 in h8.  
  rewrite inv_seg_fun_eq0_iff in h8.
  rewrite <- h5 in h8 at 1; subst.
  contradict hnin.
  rewrite in_seg_fun_iff.
  exists (pred (pred n)).
  ex_goal.
  eapply lt_pred_n.
  apply h7.
  exists hex.
  rewrite remove_im_seg_fun_ge; auto.
  gen0.
  rewrite <- (S_pred _ _ h7).
  intro; f_equal; apply proof_irrelevance.
  apply le_dec' in h8.
  destruct h8 as [h8|h8].
  specialize (h3 (pred (pred n)) (pred j)).
  hfirst h3.
  eapply lt_pred_n.
  apply h4.
  specialize (h3 hf).
  hfirst h3. 
  eapply lt_trans.
  eapply lt_pred_n.
  apply h10.
  assumption.
  specialize (h3 hf0).
  hfirst h3.
  rewrite remove_im_seg_fun_ge; auto.
  rewrite remove_im_seg_fun_ge; auto.
  gen0.
  rewrite <- (S_pred _ _ h7).
  intro.
  symmetry.
  gen0.
  rewrite <- (S_pred _ _ h10).
  intro.
  erewrite f_equal_dep.
  rewrite <- h5 at 1.
  f_equal.
  apply proof_irrelevance.
  reflexivity.
  specialize (h3 hf1).
  pose proof (f_equal S h3) as h11.
  rewrite <- (S_pred _ _ h4), <- (S_pred _ _ h10) in h11; auto.
  subst.
  rewrite inv_seg_fun_compat in h5; subst.
  symmetry.
  rewrite inv_seg_fun_eq_same_iff.
  intros j h12 h13.
  intro h14.
  contradict hnin.
  rewrite in_seg_fun_iff.
  exists (pred (pred n)).
  ex_goal.
  eapply lt_pred_n.
  apply h12.
  exists hex.
  rewrite remove_im_seg_fun_ge; auto.
  gen0.
  rewrite <- (S_pred _ _ h4).
  intro; f_equal.
  apply proof_irrelevance.
  assumption.
  rewrite inv_seg_fun_eq_iff in h9.
  destruct h9 as [h9 | h9].
  destruct h9; contradiction. 
  destruct h9 as [h9 [h10 h11]]; subst.
  specialize (h11 j hj h7).
  rewrite <- h5 in h11.
  contradict h11; f_equal; apply proof_irrelevance.
  subst; auto.
Qed.

